import asyncio
import json
import logging
import time
import traceback
from collections.abc import AsyncIterator
from dataclasses import dataclass, field
from typing import List, Optional, Tuple, Union, cast

import anchorpy
import based58
import websockets
import websockets.exceptions  # force eager imports
import websockets.legacy.client  # force eager imports
from anchorpy import Event, Provider, Wallet
from anchorpy.provider import DEFAULT_OPTIONS
from construct import Container
from jito_searcher_client import get_async_searcher_client  # type: ignore
from jito_searcher_client.generated.bundle_pb2 import Bundle  # type: ignore
from jito_searcher_client.generated.packet_pb2 import Meta, Packet  # type: ignore
from jito_searcher_client.generated.searcher_pb2 import (
    SendBundleRequest,  # type: ignore
)
from jsonrpcclient import request
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment, Confirmed
from solana.rpc.core import RPCException
from solana.rpc.types import TxOpts
from solana.rpc.websocket_api import connect
from solders.compute_budget import set_compute_unit_price
from solders.instruction import Instruction
from solders.message import MessageV0
from solders.pubkey import Pubkey
from solders.rpc.config import RpcTransactionLogsFilterMentions
from solders.system_program import TransferParams, transfer
from solders.transaction import VersionedTransaction

from zetamarkets_py import constants, pda, utils
from zetamarkets_py.events import (
    ApplyFundingEvent,
    CancelOrderEvent,
    EventMeta,
    LiquidationEvent,
    OrderCompleteEvent,
    PlaceMultiOrdersEvent,
    PlaceMultiOrdersEventWithArgs,
    PlaceOrderEvent,
    PlaceOrderEventWithArgs,
    TradeEvent,
    ZetaEnrichedEvent,
    ZetaEvent,
)
from zetamarkets_py.exchange import Exchange
from zetamarkets_py.orderbook import Orderbook
from zetamarkets_py.risk import AccountRiskSummary, Position
from zetamarkets_py.serum_client.accounts.orderbook import OrderbookAccount
from zetamarkets_py.solana_client.accounts.clock import CLOCK, Clock
from zetamarkets_py.types import (
    Asset,
    MultiOrderArgs,
    Network,
    OrderArgs,
    OrderCompleteType,
    OrderOptions,
    OrderType,
    Side,
)
from zetamarkets_py.zeta_client.accounts.cross_margin_account import CrossMarginAccount
from zetamarkets_py.zeta_client.accounts.pricing import Pricing
from zetamarkets_py.zeta_client.errors import from_tx_error
from zetamarkets_py.zeta_client.instructions import (
    cancel_all_market_orders,
    cancel_order,
    cancel_order_by_client_order_id,
    deposit_v2,
    initialize_cross_margin_account,
    initialize_cross_margin_account_manager_v2,
    initialize_open_orders_v3,
    place_multi_orders,
    place_perp_order_v5,
    withdraw_v2,
)
from zetamarkets_py.zeta_client.types import OrderArgs as ProgramOrderArgs

# TODO: add docstrings for most methods
# TODO: implement withdraw and liquidation
# TODO: implement priority fees to exchange


@dataclass
class Client:
    """
    This class represents the Zeta Client. It contains all the necessary attributes and methods
    for interacting with the Zeta Exchange from the user side, including loading the client, fetching state and pricing,
    and handling various market assets.

    Note:
        Loading the client is asynchronous, so it is recommended to use :func:`load` to
        initialize the client.
    """

    """Main provider to use for primary tx sending and fetching"""
    provider: Provider
    """Extra list of providers to use to 'double-down' for orders, sending the same transaction down multiple RPCs"""
    double_down_providers: List[Provider]

    """The network and wallet context to send transactions paid for and signed by the provider."""
    network: Network
    """The Solana network identifier (i.e. mainnet_beta, devnet etc.)."""
    connection: AsyncClient
    """The connection to the Solana network."""
    endpoint: str
    "The http(s) RPC endpoint."
    ws_endpoint: str
    """The websocket RPC endpoint."""
    exchange: Exchange
    """The Zeta Exchange object."""
    blockhash_cache: Optional[utils.BlockhashCache]
    """The blockhash cache, used when sending transactions"""

    margin_account: Optional[CrossMarginAccount]
    """The margin account of the Zeta program."""

    _margin_account_address: Optional[Pubkey]
    _open_orders_addresses: Optional[dict[Asset, Pubkey]]
    _margin_account_manager_address: Optional[Pubkey]
    _user_usdc_address: Optional[Pubkey]
    _combined_vault_address: Pubkey
    _combined_socialized_loss_address: Pubkey
    _logger: logging.Logger
    _account_exists_cache: dict[Pubkey, bool] = field(default_factory=dict)

    """Whether to also send your tx via Jito"""
    double_down_jito: bool = False
    jito_tip: int = 10000

    @classmethod
    async def load(
        cls,
        endpoint: Optional[str] = None,
        ws_endpoint: Optional[str] = None,
        double_down_endpoints: Optional[str] = None,
        commitment: Commitment = Confirmed,
        wallet: Optional[Wallet] = None,
        assets: list[Asset] = Asset.all(),
        tx_opts: TxOpts = DEFAULT_OPTIONS,
        network: Network = Network.MAINNET,
        log_level: int = logging.WARNING,
        blockhash_cache: Optional[utils.BlockhashCache] = None,
        delegator_pubkey: Optional[Pubkey] = None,
    ):
        """
        Asynchronously load the Zeta Client.

        Args:
            endpoint (str, optional): The http(s) RPC endpoint. Defaults to None.
            ws_endpoint (str, optional): The websocket RPC endpoint. Defaults to None.
            double_down_endpoints ([str], optional): The http(s) RPC endpoints to use as extra double-downs when sending critical instructions. Defaults to None.
            commitment (Commitment, optional): The commitment level of the Solana network. Defaults to Confirmed.
            wallet (Wallet, optional): The wallet used for transactions. Defaults to None.
            assets (list[Asset], optional): The list of assets to be used. Defaults to all available assets.
            tx_opts (TxOpts, optional): Transaction options. Defaults to DEFAULT_OPTIONS.
            network (Network, optional): The network of the Zeta program. Defaults to Network.MAINNET.
            log_level (int, optional): The level of logging. Defaults to logging.CRITICAL.
            blockhash_cache (Union[BlockhashCache, bool], optional): The blockhash cache. Disabled by default.
            delegator_pubkey (Pubkey, optional): If passing in a delegated wallet in the 'wallet' param, this
            is the delegator account itself so you can load positions/orders/balance/etc

        Returns:
            Client: An instance of the Client class.
        """
        logger = utils.create_logger(f"{__name__}.{cls.__name__}", log_level)

        if endpoint is None:
            endpoint = utils.cluster_endpoint(network)
        if ws_endpoint is None:
            ws_endpoint = utils.http_to_ws(endpoint)
        connection = AsyncClient(endpoint=endpoint, commitment=commitment)
        exchange = await Exchange.load(
            network=network,
            connection=connection,
            assets=assets,
        )
        if wallet is None:
            wallet = Wallet.dummy()
            logger.warning("Client in read-only mode, pass in `wallet` to enable transactions")
            _margin_account_manager_address = None
            _user_usdc_address = None
            _margin_account_address = None
            margin_account = None
            _open_orders_addresses = None
        else:
            key = wallet.public_key if delegator_pubkey is None else delegator_pubkey

            _margin_account_manager_address = pda.get_cross_margin_account_manager_address(exchange.program_id, key)
            _user_usdc_address = pda.get_associated_token_address(key, constants.USDC_MINT[network])
            _margin_account_address = pda.get_margin_account_address(exchange.program_id, key, 0)
            margin_account = await CrossMarginAccount.fetch(
                connection, _margin_account_address, connection.commitment, exchange.program_id
            )
            _open_orders_addresses = {}
            for asset in assets:
                open_orders_address = pda.get_open_orders_address(
                    exchange.program_id,
                    constants.MATCHING_ENGINE_PID[network],
                    exchange.markets[asset].address,
                    _margin_account_address,
                )
                _open_orders_addresses[asset] = open_orders_address
        provider = Provider(
            connection,
            wallet,
            tx_opts,
        )
        double_down_providers = []
        if double_down_endpoints:
            for endpoint in double_down_endpoints:
                double_down_conn = AsyncClient(endpoint=endpoint, commitment=commitment)
                double_down_providers.append(
                    Provider(
                        double_down_conn,
                        wallet,
                        tx_opts,
                    )
                )

        # additional addresses to cache
        _combined_vault_address = pda.get_combined_vault_address(exchange.program_id)
        _combined_socialized_loss_address = pda.get_combined_socialized_loss_address(exchange.program_id)

        return cls(
            provider,
            double_down_providers,
            network,
            connection,
            endpoint,
            ws_endpoint,
            exchange,
            blockhash_cache,
            margin_account,
            _margin_account_address,
            _open_orders_addresses,
            _margin_account_manager_address,
            _user_usdc_address,
            _combined_vault_address,
            _combined_socialized_loss_address,
            logger,
        )

    async def _check_user_usdc_account_exists(self):
        """
        Check if the user's USDC account exists.

        Returns:
            bool: True if the account exists, False otherwise.
        """
        if self._user_usdc_address in self._account_exists_cache:
            return self._account_exists_cache[self._user_usdc_address]
        resp = await self.connection.get_account_info(self._user_usdc_address)
        exists = resp.value is not None
        if exists:
            self._account_exists_cache[self._user_usdc_address] = exists
        return exists

    async def _check_margin_account_manager_exists(self):
        """
        Check if the margin account manager exists.

        Returns:
            bool: True if the account manager exists, False otherwise.
        """
        if self._margin_account_manager_address in self._account_exists_cache:
            return self._account_exists_cache[self._margin_account_manager_address]
        resp = await self.connection.get_account_info(self._margin_account_address)
        exists = resp.value is not None
        if exists:
            self._account_exists_cache[self._margin_account_manager_address] = exists
        return exists

    async def _check_margin_account_exists(self):
        """
        Check if the margin account exists.

        Returns:
            bool: True if the account exists, False otherwise.
        """
        if self._margin_account_address in self._account_exists_cache:
            return self._account_exists_cache[self._margin_account_address]
        account = await CrossMarginAccount.fetch(self.connection, self._margin_account_address)
        exists = account is not None
        if exists:
            self._account_exists_cache[self._margin_account_address] = exists
            self.margin_account = account
        return exists

    async def _check_open_orders_account_exists(self, asset: Asset):
        """
        Check if the open orders account for a specific asset exists.

        Args:
            asset (Asset): The asset for which to check the open orders account.

        Raises:
            Exception: If the open orders accounts are not loaded.

        Returns:
            bool: True if the open orders account exists, False otherwise.
        """
        if self._open_orders_addresses is None:
            raise Exception("Open orders accounts not loaded, cannot check if account exists")
        open_orders_address = self._open_orders_addresses[asset]
        if open_orders_address in self._account_exists_cache:
            return self._account_exists_cache[open_orders_address]
        resp = await self.connection.get_account_info(open_orders_address)
        exists = resp.value is not None
        if exists:
            self._account_exists_cache[open_orders_address] = exists
        return exists

    async def fetch_margin_state(self):
        """
        Fetch the state of the margin account.

        Raises:
            Exception: If the margin account is not loaded or not found.

        Returns:
            Tuple[Decimal, Dict[Asset, zetamarkets_py.types.Position]]: The balance and positions of the margin account.
        """
        if self.margin_account is None or self._margin_account_address is None:
            raise Exception("Margin account not loaded, cannot fetch margin account state")
        margin_account = await self.margin_account.fetch(
            self.connection, self._margin_account_address, program_id=self.exchange.program_id
        )
        if margin_account is None:
            raise Exception("Margin account not found, cannot fetch margin state")
        balance = utils.convert_fixed_int_to_decimal(margin_account.balance)
        positions = {
            Asset.from_index(i): Position.from_margin_account(margin_account, i)
            for i in range(len(self.exchange.assets))
        }
        return balance, positions

    async def get_account_risk_summary(self):
        """
        Get the risk summary of the account.

        This method fetches the margin state of the account, calculates the equity, margin parameters, margin usage,
        and leverage, and returns an AccountRiskSummary object.

        Returns:
            AccountRiskSummary: The risk summary of the account.
        """
        # Batched RPC call to margin acc and prices
        account_infos = await anchorpy.utils.rpc.get_multiple_accounts(
            self.connection, [self._margin_account_address, self.exchange._pricing_address]
        )
        margin_account = CrossMarginAccount.decode(account_infos[0].account.data)
        pricing_account = Pricing.decode(account_infos[1].account.data)
        return AccountRiskSummary.from_margin_and_pricing_accounts(margin_account, pricing_account)

    async def fetch_open_orders(self, asset: Asset):
        """
        Fetch the open orders for a specific asset.

        Args:
            asset (Asset): The asset for which to fetch the open orders.

        Raises:
            Exception: If the open orders accounts are not loaded.

        Returns:
            List[Order]: The open orders for the asset.
        """
        if self._open_orders_addresses is None:
            raise Exception("Open orders accounts not loaded, cannot fetch open orders")
        oo = await self.exchange.markets[asset].load_orders_for_owner(self._open_orders_addresses[asset])
        return oo

    async def fetch_clock(self):
        """
        Fetch the Solana clock.

        Raises:
            Exception: If the clock is not found.

        Returns:
            Clock: The clock.
        """
        clock = await Clock.fetch(self.connection)
        if clock is None:
            raise Exception("Clock not found, cannot fetch clock")
        return clock

    async def _account_subscribe(
        self,
        address: Pubkey,
        commitment: Commitment,
        encoding: str = "base64+zstd",
    ) -> AsyncIterator[Tuple[bytes, int]]:
        """
        Subscribe to an account and yield account data and slot.

        Args:
            address (Pubkey): The public key of the account to subscribe to.
            commitment (Commitment): The commitment level to use for the subscription.
            encoding (str, optional): The encoding to use for the subscription. Defaults to "base64+zstd".

        Yields:
            AsyncIterator[Tuple[bytes, int]]: An async iterator that yields tuples of account data and slot.
        """
        async for ws in connect(self.ws_endpoint):
            try:
                await ws.account_subscribe(  # type: ignore
                    address,
                    commitment=commitment,
                    encoding=encoding,
                )
                first_resp = await ws.recv()
                subscription_id = cast(int, first_resp[0].result)  # type: ignore
                async for msg in ws:
                    try:
                        slot = int(msg[0].result.context.slot)  # type: ignore
                        account_bytes = cast(bytes, msg[0].result.value.data)  # type: ignore
                        yield account_bytes, slot
                    except Exception:
                        self._logger.error(f"Error processing account data: {traceback.format_exc()}")
                        break
                await ws.account_unsubscribe(subscription_id)  # type: ignore
            except websockets.exceptions.ConnectionClosed:
                self._logger.warning("Websocket closed, reconnecting...")
                continue

    async def subscribe_orderbook(
        self, asset: Asset, side: Side, commitment: Optional[Commitment] = None
    ) -> AsyncIterator[Tuple[Orderbook, int]]:
        """
        Subscribe to an orderbook and yield orderbook data and slot.

        Args:
            asset (Asset): The asset for which to subscribe to the orderbook.
            side (Side): The side of the orderbook to subscribe to.
            commitment (Commitment, optional): The commitment level to use for the subscription. Defaults to None.

        Yields:
            AsyncIterator[Tuple[Orderbook, int]]: An async iterator that yields tuples of orderbook data and slot.
        """
        commitment = commitment or self.connection.commitment
        address = (
            self.exchange.markets[asset]._market_state.bids
            if side == Side.Bid
            else self.exchange.markets[asset]._market_state.asks
        )

        self._logger.info(f"Subscribing to Orderbook:{side}.")
        async for account_bytes, slot in self._account_subscribe(address, commitment):
            account = OrderbookAccount.decode(account_bytes)
            orderbook = Orderbook(side, account, self.exchange.markets[asset]._market_state)
            yield orderbook, slot

    async def subscribe_clock(self, commitment: Optional[Commitment] = None) -> AsyncIterator[Tuple[Clock, int]]:
        """
        Subscribe to a clock and yield clock data and slot.

        Args:
            commitment (Commitment, optional): The commitment level to use for the subscription. Defaults to None.

        Yields:
            AsyncIterator[Tuple[Clock, int]]: An async iterator that yields tuples of clock data and slot.
        """
        commitment = commitment or self.connection.commitment

        self._logger.info("Subscribing to Clock")
        async for account_bytes, slot in self._account_subscribe(CLOCK, commitment):
            clock = Clock.decode(account_bytes)
            yield clock, slot

    async def subscribe_events(
        self, commitment: Optional[Commitment] = None, ignore_third_party_events: bool = True
    ) -> AsyncIterator[Tuple[List[ZetaEvent], EventMeta]]:
        """
        Subscribe to events and yield event data and slot.

        Args:
            commitment (Commitment, optional): The commitment level to use for the subscription. Defaults to None.

        Yields:
            AsyncIterator[Tuple[List[ZetaEvent], int]]: An async iterator that yields tuples of event data and slot.
        """
        if ignore_third_party_events:
            if self._margin_account_address is None:
                raise Exception("Margin account not loaded, cannot subscribe to events")
            pubkey = self._margin_account_address
        else:
            pubkey = self.exchange.program_id
        commitment = commitment or self.connection.commitment
        async for ws in connect(self.ws_endpoint):
            try:
                # Subscribe to logs that mention the margin account
                await ws.logs_subscribe(  # type: ignore
                    commitment=commitment,
                    filter_=RpcTransactionLogsFilterMentions(pubkey),
                )
                first_resp = await ws.recv()
                subscription_id = cast(int, first_resp[0].result)  # type: ignore
                async for msg in ws:
                    try:
                        events, meta = self._parse_event_payload(msg, ignore_third_party_events)
                        if len(events) > 0 or not meta.is_successful:
                            yield events, meta

                    except Exception:
                        self._logger.error(f"Error processing event data: {traceback.format_exc()}")
                        break
                await ws.logs_unsubscribe(subscription_id)  # type: ignore
            except websockets.exceptions.ConnectionClosed:
                self._logger.warning("Websocket closed, reconnecting...")
                continue

    def _parse_event_payload(self, msg, ignore_third_party_events: bool = True) -> Tuple[List[ZetaEvent], EventMeta]:
        """
        Parse the event payload from the message.

        Args:
            msg: The message received from the websocket.

        Returns:
            Tuple[List[ZetaEvent], EventMeta]: A tuple containing a list of ZetaEvents and event metadata.
        """
        slot = int(msg[0].result.context.slot)
        logs = cast(list[str], msg[0].result.value.logs)  # type: ignore
        error = msg[0].result.value.err
        signature = msg[0].result.value.signature
        meta = EventMeta(slot, error, signature)
        if error is not None:
            return [], meta

        parsed: list[Event] = []
        self.exchange._event_parser.parse_logs(logs, lambda evt: parsed.append(evt))
        events = []
        for event in parsed:
            if event.name.startswith(PlaceOrderEvent.__name__):
                place_order_event = PlaceOrderEvent.from_event(event)
                if not ignore_third_party_events or place_order_event.margin_account == self._margin_account_address:
                    events.append(place_order_event)
            elif event.name.startswith(OrderCompleteEvent.__name__):
                order_complete_event = OrderCompleteEvent.from_event(event)
                # Ignore fills
                if (
                    order_complete_event.order_complete_type == OrderCompleteType.Cancel
                    or order_complete_event.order_complete_type == OrderCompleteType.Booted
                ):
                    cancel_event = CancelOrderEvent.from_order_complete_event(order_complete_event)
                    if not ignore_third_party_events or cancel_event.margin_account == self._margin_account_address:
                        events.append(cancel_event)
            elif event.name.startswith(TradeEvent.__name__):
                trade_event = TradeEvent.from_event(event)
                if not ignore_third_party_events or trade_event.margin_account == self._margin_account_address:
                    events.append(trade_event)
            elif event.name.startswith(LiquidationEvent.__name__):
                liquidation_event = LiquidationEvent.from_event(event)
                if (
                    not ignore_third_party_events
                    or liquidation_event.liquidatee_margin_account == self._margin_account_address
                ):
                    events.append(liquidation_event)
            elif event.name.startswith(ApplyFundingEvent.__name__):
                funding_event = ApplyFundingEvent.from_event(event)
                if not ignore_third_party_events or funding_event.margin_account == self._margin_account_address:
                    events.append(funding_event)
            else:
                pass

        return events, meta

    async def subscribe_transactions(
        self,
        commitment: Optional[Commitment] = None,
        ignore_truncation: bool = False,
        ignore_third_party_transactions: bool = True,
    ) -> AsyncIterator[Tuple[List[ZetaEnrichedEvent], EventMeta]]:
        """
        This method is used to subscribe to transactions.

        Args:
            commitment (Optional[Commitment], optional): The commitment level to use for the subscription.
                Defaults to None.
            ignore_truncation(bool): Bool to ignore the "Logs truncated, missing event data" warning.
                Defaults to False.

        Yields:
            Tuple[List[ZetaEnrichedEvent], EventMeta]: A tuple containing a list of ZetaEnrichedEvents and event metadata.

        Warning:
            This method is experimental and requires a Triton RPC node.
        """

        if "rpcpool.com" not in self.ws_endpoint:
            self._logger.warning(
                'Provided ws_endpoint does not contain "rpcpool.com". This method is experimental and'
                " requires a Triton RPC node for transactionSubscribe"
            )

        commitment = commitment or self.connection.commitment
        # TODO: upgrade to websockets 12.0
        # TODO: modify solanapy websocket stuff and make it support txs + types and subclassing
        # (so we dont have to handle json)
        async for ws in websockets.legacy.client.connect(self.ws_endpoint + "/whirligig"):
            try:
                transaction_subscribe = request(
                    "transactionSubscribe",
                    params=[
                        {
                            "mentions": [
                                (
                                    str(self._margin_account_address)
                                    if ignore_third_party_transactions
                                    else str(self.exchange.program_id)
                                )
                            ],
                            # "failed": False,
                            "vote": False,
                        },
                        {
                            "commitment": str(commitment),
                        },
                    ],
                )
                await ws.send(json.dumps(transaction_subscribe))

                first_resp = await ws.recv()
                subscription_id = cast(int, first_resp)

                async for msg in ws:
                    try:
                        events, meta = self._parse_transaction_payload(
                            msg, ignore_truncation, ignore_third_party_transactions
                        )
                        if len(events) > 0 or not meta.is_successful:
                            yield events, meta
                    except Exception:
                        self._logger.error(f"Error processing transaction data: {traceback.format_exc()}")
                        break

                transaction_unsubscribe = request(
                    "transactionUnsubscribe",
                    params=[subscription_id],
                )
                await ws.send(json.dumps(transaction_unsubscribe))
            except websockets.exceptions.ConnectionClosed:
                self._logger.warning("Websocket closed, reconnecting...")
                continue

    def _parse_transaction_payload(
        self, msg, ignore_truncation: bool = False, ignore_third_party_transactions: bool = True
    ) -> Tuple[List[ZetaEnrichedEvent], EventMeta]:
        """
        Parse the transaction payload from the message.

        Args:
            msg: The message received from the websocket.
            ignore_truncation: Bool to ignore the "Logs truncated, missing event data" warning. Defaults to False.
            ignore_third_party_transactions: Bool to ignore transactions from other users, filtering by margin account address

        Returns:
            Tuple[List[ZetaEnrichedEvent], EventMeta]: A tuple containing a list of ZetaEnrichedEvent and event metadata.
        """
        json_msg = json.loads(msg)
        slot = int(json_msg["params"]["result"]["context"]["slot"])
        tx_value = json_msg["params"]["result"]["value"]
        log_messages = tx_value["meta"]["logMessages"]
        message = tx_value["transaction"]["message"]
        error = tx_value["meta"]["err"]
        signature = str(based58.b58encode(bytes(tx_value["signature"])), encoding="utf-8")
        meta = EventMeta(slot, error, signature)
        if error is not None:
            return [], meta

        if isinstance(message[0], int) or "instructions" not in message[0]:
            message_indexed = message[1]
        else:
            message_indexed = message[0]

        ixs = message_indexed["instructions"][1:]
        ix_args: list[Optional[Container]] = []
        ix_names: list[Optional[str]] = []
        events_to_return = []

        for ix in ixs:
            acc_keys_raw = message_indexed["accountKeys"][1:]
            account_keys = [str(based58.b58encode(bytes(a)), encoding="utf-8") for a in acc_keys_raw]
            loaded_addresses = tx_value["meta"]["loadedAddresses"]
            loaded_addresses_list = account_keys + loaded_addresses["writable"] + loaded_addresses["readonly"]
            owner_address = loaded_addresses_list[ix["programIdIndex"]]

            if owner_address != str(constants.ZETA_PID[self.network]):
                ix_args.append(None)
                ix_names.append(None)
                continue
            data = self.exchange.program.coder.instruction.parse(bytes(ix["data"][1:]))
            ix_args.append(data.data)
            ix_names.append(data.name)

        # Split log_messages every time we see "invoke [1]"
        split_log_messages = []
        split_indices = []
        for i in range(len(log_messages)):
            if log_messages[i] == "Log truncated":
                if ignore_truncation:
                    break
                self._logger.warning("Logs truncated, missing event data")
            if log_messages[i].endswith("invoke [1]"):
                split_indices.append(i)

        split_log_messages = [log_messages[i:j] for i, j in zip([0] + split_indices, split_indices + [None])]
        if len(split_log_messages) > 0:
            split_log_messages = split_log_messages[1:]

        if not ignore_truncation and (
            len(ix_args) != len(split_log_messages) or len(ix_names) != len(split_log_messages)
        ):
            raise Exception("Mismatched transaction info lengths")

        # For each individual instruction, find the ix name and the events
        for i in range(len(split_log_messages)):
            # First log line will always be "...invoke [1]", second will be "Program log: Instruction: <ix_name>"
            ix_name = ix_names[i]
            ix_arg = ix_args[i]

            if ix_name is None or ix_arg is None:
                continue

            chunk = split_log_messages[i]
            events: list[Event] = []
            self.exchange._event_parser.parse_logs(chunk, lambda evt: events.append(evt))

            # Depending on the instruction and event we can get different data from the args
            for event in events:
                # Skip event that aren't for our account but mention our account
                # eg if we do a taker trade, we want to skip the maker crank events
                if ignore_third_party_transactions:
                    account_check = (
                        str(event.data.liquidatee_margin_account)
                        if ix_name.startswith("liquidation")
                        else str(event.data.margin_account)
                    )
                    if account_check != str(self._margin_account_address):
                        continue

                if ix_name.startswith("place_perp_order"):
                    if event.name.startswith(TradeEvent.__name__):
                        events_to_return.append(TradeEvent.from_event(event))  # Taker fill
                    elif event.name.startswith(PlaceOrderEvent.__name__):
                        events_to_return.append(PlaceOrderEventWithArgs.from_event_and_args(event, ix_arg))

                elif ix_name.startswith("place_multi_orders"):
                    if event.name.startswith(PlaceMultiOrdersEvent.__name__):
                        events_to_return.append(PlaceMultiOrdersEventWithArgs.from_event_and_args(event, ix_arg))

                elif ix_name.startswith("crank_event_queue"):
                    if event.name.startswith(TradeEvent.__name__):
                        events_to_return.append(TradeEvent.from_event(event))  # Maker fill\
                    elif str(event.name).startswith(str(OrderCompleteEvent.__name__)):
                        order_complete_event = OrderCompleteEvent.from_event(event)
                        if order_complete_event.order_complete_type == OrderCompleteType.Booted:  # TIF expiry
                            events_to_return.append(CancelOrderEvent.from_order_complete_event(order_complete_event))

                elif ix_name.startswith("cancel_"):
                    if event.name.startswith(OrderCompleteEvent.__name__):
                        order_complete_event = OrderCompleteEvent.from_event(event)
                        if (
                            order_complete_event.order_complete_type == OrderCompleteType.Cancel
                            or order_complete_event.order_complete_type == OrderCompleteType.Booted
                        ):
                            events_to_return.append(CancelOrderEvent.from_order_complete_event(order_complete_event))

                elif ix_name.startswith("apply_perp_funding"):
                    if event.name.startswith(ApplyFundingEvent.__name__):
                        events_to_return.append(ApplyFundingEvent.from_event(event))

        return events_to_return, meta

    # Instructions

    async def deposit(self, amount: float, subaccount_index: int = 0, priority_fee: int = 0):
        """
        This method is used to deposit a specified amount into the user's margin account.

        Args:
            amount (float): The amount to be deposited.
            subaccount_index (int, optional): The index of the subaccount. Defaults to 0.
            priority_fee (int): Additional priority fee, in microlamports per CU. Defaults to 0.

        Raises:
            Exception: If the user does not have a USDC account.

        Returns:
            Transaction: The transaction object of the deposit operation.
        """
        ixs = []
        if priority_fee > 0:
            ixs.append(set_compute_unit_price(priority_fee))
        if not await self._check_margin_account_manager_exists():
            self._logger.info("User has no cross-margin account manager, creating one...")
            ixs.append(self._init_margin_account_manager_ix())
        # Create margin account if user doesn't have one
        if not await self._check_margin_account_exists():
            self._logger.info("User has no cross-margin account, creating one...")
            ixs.append(self._init_margin_account_ix(subaccount_index))
        # Check they have an existing USDC account
        if await self._check_user_usdc_account_exists():
            ixs.append(self._deposit_ix(amount))
        else:
            raise Exception("User has no USDC, cannot deposit to margin account")

        self._logger.info(f"Depositing {amount} USDC to margin account")
        return await self._send_versioned_transaction(ixs)

    def _init_margin_account_manager_ix(self) -> Instruction:
        """
        Initialize the margin account manager instruction.

        Raises:
            Exception: If the margin account manager address is not loaded.

        Returns:
            Instruction: The initialized cross margin account manager instruction.
        """
        if self._margin_account_manager_address is None:
            raise Exception("Margin account manager address not loaded, cannot deposit")
        return initialize_cross_margin_account_manager_v2(
            {"referrer": constants.MAINTAINER_REFERRAL},
            {
                "cross_margin_account_manager": self._margin_account_manager_address,
                "authority": self.provider.wallet.public_key,
                "payer": self.provider.wallet.public_key,
                "zeta_program": self.exchange.program_id,
            },
            self.exchange.program_id,
        )

    def _init_margin_account_ix(self, subaccount_index: int = 0) -> Instruction:
        """
        Initialize the margin account instruction.

        Args:
            subaccount_index (int, optional): The index of the subaccount. Defaults to 0.

        Raises:
            Exception: If the margin account address is not loaded.

        Returns:
            Instruction: The initialized cross margin account instruction.
        """
        if self._margin_account_address is None or self._margin_account_manager_address is None:
            raise Exception("Margin account address not loaded, cannot deposit")
        return initialize_cross_margin_account(
            {"subaccount_index": subaccount_index},
            {
                "cross_margin_account": self._margin_account_address,
                "cross_margin_account_manager": self._margin_account_manager_address,
                "authority": self.provider.wallet.public_key,
                "payer": self.provider.wallet.public_key,
                "zeta_program": self.exchange.program_id,
            },
            self.exchange.program_id,
        )

    def _deposit_ix(self, amount: float) -> Instruction:
        """
        Deposit instruction.

        Args:
            amount (float): The amount to be deposited.

        Raises:
            Exception: If the user USDC address is not loaded.

        Returns:
            Instruction: The deposit instruction.
        """
        if self._user_usdc_address is None or self._margin_account_address is None:
            raise Exception("User USDC address not loaded, cannot deposit")
        return deposit_v2(
            {"amount": utils.convert_decimal_to_fixed_int(amount, constants.MIN_NATIVE_TICK_SIZE)},
            {
                "margin_account": self._margin_account_address,
                "vault": self._combined_vault_address,
                "user_token_account": self._user_usdc_address,
                "socialized_loss_account": self._combined_socialized_loss_address,
                "authority": self.provider.wallet.public_key,
                "state": self.exchange._state_address,
                "pricing": self.exchange._pricing_address,
            },
            self.exchange.program_id,
        )

    # TODO: withdraw (and optionally close)
    async def withdraw(self, amount: float, priority_fee: int = 0):
        """Initiates a withdrawal of a specified amount with an optional priority fee.

        Args:
            amount (float): The amount to withdraw.
            priority_fee (int): Additional priority fee, in microlamports per CU. Defaults to 0.

        Raises:
            Exception: If the user does not have a USDC account.

        Returns:
            Transaction: The transaction object of the withdrawal operation.

        """
        ixs = []
        if priority_fee > 0:
            ixs.append(set_compute_unit_price(priority_fee))
        # Check they have an existing USDC account
        if await self._check_user_usdc_account_exists():
            ixs.append(self._withdraw_ix(amount))
        else:
            raise Exception("User has no USDC, cannot withdraw from margin account")

        self._logger.info(f"Withdrawing {amount} USDC from margin account")
        return await self._send_versioned_transaction(ixs)

    def _withdraw_ix(self, amount: float) -> Instruction:
        """
        Withdraw instruction.

        Args:
            amount (float): The amount to be withdrawn.

        Raises:
            Exception: If the user USDC address is not loaded.

        Returns:
            Instruction: The withdraw instruction.
        """
        if self._user_usdc_address is None or self._margin_account_address is None:
            raise Exception("User USDC address not loaded, cannot withdraw")
        return withdraw_v2(
            {"amount": utils.convert_decimal_to_fixed_int(amount, constants.MIN_NATIVE_TICK_SIZE)},
            {
                "margin_account": self._margin_account_address,
                "vault": self._combined_vault_address,
                "user_token_account": self._user_usdc_address,
                "socialized_loss_account": self._combined_socialized_loss_address,
                "authority": self.provider.wallet.public_key,
                "state": self.exchange._state_address,
                "pricing": self.exchange._pricing_address,
            },
            self.exchange.program_id,
        )

    def _init_open_orders_ix(self, asset: Asset) -> Instruction:
        """
        Initialize the open orders instruction.

        Args:
            asset (Asset): The asset for which to initialize the open orders.

        Raises:
            Exception: If the asset is not loaded into the client or if the open orders address is not loaded.

        Returns:
            Instruction: The initialized open orders instruction.
        """
        if asset not in self.exchange.assets:
            raise Exception(f"Asset {asset.name} not loaded into client, cannot initialize open orders")
        if self._open_orders_addresses is None or self._margin_account_address is None:
            raise Exception("Open orders address not loaded, cannot place order")
        return initialize_open_orders_v3(
            {"asset": asset.to_program_type()},
            {
                "state": self.exchange._state_address,
                "dex_program": constants.MATCHING_ENGINE_PID[self.network],
                "open_orders": self._open_orders_addresses[asset],
                "cross_margin_account": self._margin_account_address,
                "authority": self.provider.wallet.public_key,
                "payer": self.provider.wallet.public_key,
                "market": self.exchange.markets[asset].address,
                "serum_authority": self.exchange._serum_authority_address,
                "open_orders_map": pda.get_open_orders_map_address(
                    self.exchange.program_id, self._open_orders_addresses[asset]
                ),
            },
            self.exchange.program_id,
        )

    def _place_order_ix(
        self,
        asset: Asset,
        price: float,
        size: float,
        side: Side,
        order_opts: Optional[OrderOptions] = None,
        tif_buffer: int = 0,
    ) -> Instruction:
        """
        Build a PlaceOrder instruction.

        Args:
            asset (Asset): The asset for which to place the order.
            price (float): The price of the order.
            size (float): The size of the order.
            side (Side): The side of the order (bid or ask).
            order_opts (Optional[OrderOptions], optional): The options for the order. Defaults to None.

        Raises:
            Exception: If the asset is not loaded into the client, or if the margin account address or open orders
                addresses are not loaded.

        Returns:
            Instruction: The place order instruction.
        """
        if order_opts is None:
            order_opts = OrderOptions()

        unix_timestamp = int(time.time())
        tif_offset = (
            utils.get_tif_offset(
                order_opts.expiry_ts,
                self.exchange.markets[asset]._market_state.epoch_length,
                unix_timestamp,  # self.exchange.clock.account.unix_timestamp,
                tif_buffer,
            )
            if order_opts.expiry_ts
            else None
        )

        if asset not in self.exchange.assets:
            raise Exception(f"Asset {asset.name} not loaded into client, cannot place order")
        if self._margin_account_address is None:
            raise Exception("Margin account address not loaded, cannot place order")
        if self._open_orders_addresses is None:
            raise Exception("Open orders addresses not loaded, cannot place order")

        return place_perp_order_v5(
            {
                "price": utils.convert_decimal_to_fixed_int(
                    price, utils.get_fixed_tick_size(self.exchange.state, asset)
                ),
                "size": utils.convert_decimal_to_fixed_lot(
                    size, utils.get_fixed_min_lot_size(self.exchange.state, asset)
                ),
                "side": side.to_program_type(),
                "order_type": order_opts.order_type.to_program_type(),
                "reduce_only": False,
                "client_order_id": order_opts.client_order_id,
                "tif_offset": tif_offset,
                "tag": order_opts.tag,
                "asset": asset.to_program_type(),
                "self_trade_behavior": (
                    order_opts.self_trade_behavior.to_program_type()
                    if order_opts.self_trade_behavior is not None
                    else None
                ),
            },
            {
                "authority": self.provider.wallet.public_key,
                "place_order_accounts": {
                    "state": self.exchange._state_address,
                    "pricing": self.exchange._pricing_address,
                    "margin_account": self._margin_account_address,
                    "dex_program": constants.MATCHING_ENGINE_PID[self.network],
                    "serum_authority": self.exchange._serum_authority_address,
                    "open_orders": self._open_orders_addresses[asset],
                    "market_accounts": {
                        "market": self.exchange.markets[asset].address,
                        "request_queue": self.exchange.markets[asset]._market_state.request_queue,
                        "event_queue": self.exchange.markets[asset]._market_state.event_queue,
                        "bids": self.exchange.markets[asset]._market_state.bids,
                        "asks": self.exchange.markets[asset]._market_state.asks,
                        "coin_vault": self.exchange.markets[asset]._market_state.base_vault,
                        "pc_vault": self.exchange.markets[asset]._market_state.quote_vault,
                        "order_payer_token_account": (
                            self.exchange.markets[asset]._quote_zeta_vault_address
                            if side == Side.Bid
                            else self.exchange.markets[asset]._base_zeta_vault_address
                        ),
                        "coin_wallet": self.exchange.markets[asset]._base_zeta_vault_address,
                        "pc_wallet": self.exchange.markets[asset]._quote_zeta_vault_address,
                    },
                    "oracle": constants.PYTH_PRICE_FEEDS[self.network][asset],
                    "oracle_backup_feed": self.exchange.pricing.oracle_backup_feeds[asset.to_index()],
                    "oracle_backup_program": constants.CHAINLINK_PID,
                    "market_mint": (
                        self.exchange.markets[asset]._market_state.quote_mint
                        if side == Side.Bid
                        else self.exchange.markets[asset]._market_state.base_mint
                    ),
                    "mint_authority": self.exchange._mint_authority_address,
                    "perp_sync_queue": self.exchange.pricing.perp_sync_queues[asset.to_index()],
                },
            },
            self.exchange.program_id,
        )

    def _place_multi_orders_ix(
        self,
        asset: Asset,
        bid_orders: list[MultiOrderArgs],
        ask_orders: list[MultiOrderArgs],
        order_type: OrderType,
        tif_buffer: int = 0,
    ) -> Instruction:
        """
        Build a PlaceMultiOrders instruction.

        Args:
            asset (Asset): The asset for which to place the order.
            price (float): The price of the order.
            size (float): The size of the order.
            side (Side): The side of the order (bid or ask).
            order_opts (Optional[OrderOptions], optional): The options for the order. Defaults to None.

        Raises:
            Exception: If the asset is not loaded into the client, or if the margin account address or open orders
                addresses are not loaded.

        Returns:
            Instruction: The place order instruction.
        """
        if asset not in self.exchange.assets:
            raise Exception(f"Asset {asset.name} not loaded into client, cannot place order")
        if self._margin_account_address is None:
            raise Exception("Margin account address not loaded, cannot place order")
        if self._open_orders_addresses is None:
            raise Exception("Open orders addresses not loaded, cannot place order")

        bid_orders_program = []
        ask_orders_program = []

        unix_timestamp = int(time.time())
        for o in bid_orders:
            tif_offset = (
                utils.get_tif_offset(
                    o.expiry_ts,
                    self.exchange.markets[asset]._market_state.epoch_length,
                    unix_timestamp,  # self.exchange.clock.account.unix_timestamp,
                    tif_buffer,
                )
                if o.expiry_ts
                else None
            )

            p = utils.convert_decimal_to_fixed_int(o.price, utils.get_fixed_tick_size(self.exchange.state, asset))
            s = utils.convert_decimal_to_fixed_lot(o.size, utils.get_fixed_min_lot_size(self.exchange.state, asset))
            program_order = ProgramOrderArgs(p, s, o.client_order_id, tif_offset)
            bid_orders_program.append(program_order)

        for o in ask_orders:
            tif_offset = (
                utils.get_tif_offset(
                    o.expiry_ts,
                    self.exchange.markets[asset]._market_state.epoch_length,
                    unix_timestamp,  # self.exchange.clock.account.unix_timestamp,
                    tif_buffer,
                )
                if o.expiry_ts
                else None
            )

            p = utils.convert_decimal_to_fixed_int(o.price, utils.get_fixed_tick_size(self.exchange.state, asset))
            s = utils.convert_decimal_to_fixed_lot(o.size, utils.get_fixed_min_lot_size(self.exchange.state, asset))
            program_order = ProgramOrderArgs(p, s, o.client_order_id, tif_offset)
            ask_orders_program.append(program_order)

        return place_multi_orders(
            {
                "asset": asset.to_program_type(),
                "bid_orders": bid_orders_program,
                "ask_orders": ask_orders_program,
                "order_type": order_type.to_program_type(),
            },
            {
                "authority": self.provider.wallet.public_key,
                "state": self.exchange._state_address,
                "pricing": self.exchange._pricing_address,
                "margin_account": self._margin_account_address,
                "dex_program": constants.MATCHING_ENGINE_PID[self.network],
                "serum_authority": self.exchange._serum_authority_address,
                "open_orders": self._open_orders_addresses[asset],
                "market": self.exchange.markets[asset].address,
                "request_queue": self.exchange.markets[asset]._market_state.request_queue,
                "event_queue": self.exchange.markets[asset]._market_state.event_queue,
                "bids": self.exchange.markets[asset]._market_state.bids,
                "asks": self.exchange.markets[asset]._market_state.asks,
                "market_base_vault": self.exchange.markets[asset]._market_state.base_vault,
                "market_quote_vault": self.exchange.markets[asset]._market_state.quote_vault,
                "zeta_base_vault": self.exchange.markets[asset]._base_zeta_vault_address,
                "zeta_quote_vault": self.exchange.markets[asset]._quote_zeta_vault_address,
                "oracle": constants.PYTH_PRICE_FEEDS[self.network][asset],
                "oracle_backup_feed": self.exchange.pricing.oracle_backup_feeds[asset.to_index()],
                "oracle_backup_program": constants.CHAINLINK_PID,
                "market_base_mint": self.exchange.markets[asset]._market_state.base_mint,
                "market_quote_mint": self.exchange.markets[asset]._market_state.quote_mint,
                "mint_authority": self.exchange._mint_authority_address,
                "perp_sync_queue": self.exchange.pricing.perp_sync_queues[asset.to_index()],
            },
            self.exchange.program_id,
        )

    async def cancel_order(self, asset: Asset, order_id: int, side: Side, priority_fee: int = 0):
        """
        Cancel an order.

        Args:
            asset (Asset): The asset for which to cancel the order.
            order_id (int): The ID of the order to cancel.
            side (Side): The side of the order (buy or sell).

        Returns:
            Transaction: The transaction of the cancelled order.
        """
        ixs = []
        if priority_fee > 0:
            ixs.append(set_compute_unit_price(priority_fee))
        ixs.append(self._cancel_order_ix(asset, order_id, side))
        self._logger.info(f"Cancelling order with order id {order_id} for {asset}")
        return await self._send_versioned_transaction(ixs)

    def _cancel_order_ix(self, asset: Asset, order_id: int, side: Side) -> Instruction:
        """
        Build a CancelOrder instruction.

        Args:
            asset (Asset): The asset for which to cancel the order.
            order_id (int): The ID of the order to cancel.
            side (Side): The side of the order (bid or ask).

        Raises:
            Exception: If the margin account address or open orders addresses are not loaded.

        Returns:
            Instruction: The cancel order instruction.
        """
        if self._margin_account_address is None:
            raise Exception("Margin account address not loaded, cannot cancel order")
        if self._open_orders_addresses is None:
            raise Exception("Open orders addresses not loaded, cannot cancel order")
        return cancel_order(
            {"side": side.to_program_type(), "order_id": order_id, "asset": asset.to_program_type()},
            {
                "authority": self.provider.wallet.public_key,
                "cancel_accounts": {
                    "state": self.exchange._state_address,
                    "margin_account": self._margin_account_address,
                    "dex_program": constants.MATCHING_ENGINE_PID[self.network],
                    "serum_authority": self.exchange._serum_authority_address,
                    "open_orders": self._open_orders_addresses[asset],
                    "market": self.exchange.markets[asset].address,
                    "bids": self.exchange.markets[asset]._market_state.bids,
                    "asks": self.exchange.markets[asset]._market_state.asks,
                    "event_queue": self.exchange.markets[asset]._market_state.event_queue,
                },
            },
            self.exchange.program_id,
        )

    def _cancel_order_by_client_order_id_ix(self, asset: Asset, client_order_id: int) -> Instruction:
        if self._margin_account_address is None:
            raise Exception("Margin account address not loaded, cannot cancel order")
        if self._open_orders_addresses is None:
            raise Exception("Open orders addresses not loaded, cannot cancel order")
        return cancel_order_by_client_order_id(
            {"client_order_id": client_order_id, "asset": asset.to_program_type()},
            {
                "authority": self.provider.wallet.public_key,
                "cancel_accounts": {
                    "state": self.exchange._state_address,
                    "margin_account": self._margin_account_address,
                    "dex_program": constants.MATCHING_ENGINE_PID[self.network],
                    "serum_authority": self.exchange._serum_authority_address,
                    "open_orders": self._open_orders_addresses[asset],
                    "market": self.exchange.markets[asset].address,
                    "bids": self.exchange.markets[asset]._market_state.bids,
                    "asks": self.exchange.markets[asset]._market_state.asks,
                    "event_queue": self.exchange.markets[asset]._market_state.event_queue,
                },
            },
            self.exchange.program_id,
        )

    async def cancel_order_by_client_order_id(self, asset: Asset, client_order_id: int, priority_fee: int = 0):
        ixs = []
        if priority_fee > 0:
            ixs.append(set_compute_unit_price(priority_fee))
        ixs.append(self._cancel_order_by_client_order_id_ix(asset, client_order_id))
        self._logger.info(f"Cancelling order with client order id {client_order_id} for {asset}")
        return await self._send_versioned_transaction(ixs)

    def _cancel_orders_for_market_ix(self, asset: Asset) -> Instruction:
        """
        Build an instruction for cancelling all orders on a given market.

        Args:
            asset (Asset): The asset for which to cancel all orders.

        Raises:
            Exception: If the margin account address or open orders addresses are not loaded.

        Returns:
            Instruction: The cancel all orders for a market instruction.
        """
        if self._margin_account_address is None:
            raise Exception("Margin account address not loaded, cannot cancel orders")
        if self._open_orders_addresses is None:
            raise Exception("Open orders addresses not loaded, cannot cancel orders")
        return cancel_all_market_orders(
            {"asset": asset.to_program_type()},
            {
                "authority": self.provider.wallet.public_key,
                "cancel_accounts": {
                    "state": self.exchange._state_address,
                    "margin_account": self._margin_account_address,
                    "dex_program": constants.MATCHING_ENGINE_PID[self.network],
                    "serum_authority": self.exchange._serum_authority_address,
                    "open_orders": self._open_orders_addresses[asset],
                    "market": self.exchange.markets[asset].address,
                    "bids": self.exchange.markets[asset]._market_state.bids,
                    "asks": self.exchange.markets[asset]._market_state.asks,
                    "event_queue": self.exchange.markets[asset]._market_state.event_queue,
                },
            },
            self.exchange.program_id,
        )

    async def cancel_orders_for_market(
        self,
        asset: Asset,
        pre_instructions: Optional[list[Instruction]] = None,
        post_instructions: Optional[list[Instruction]] = None,
        priority_fee: int = 0,
    ):
        """
        Cancel all orders for a market.

        Args:
            asset (Asset): The asset for which to cancel all orders.
            pre_instructions (Optional[list[Instruction]], optional): The list of instructions to execute before
                cancelling the orders. Defaults to None.
            post_instructions (Optional[list[Instruction]], optional): The list of instructions to execute after
                cancelling the orders. Defaults to None.
            priority_fee (int): Additional priority fee, in microlamports per CU. Defaults to 0.

        Returns:
            Transaction: The transaction of the cancelled orders.
        """
        ixs = []
        if priority_fee > 0:
            ixs.append(set_compute_unit_price(priority_fee))
        if pre_instructions is not None:
            ixs.extend(pre_instructions)
        ixs.append(self._cancel_orders_for_market_ix(asset))
        if post_instructions is not None:
            ixs.extend(post_instructions)
        self._logger.info(f"Cancelling all orders for {asset}")
        return await self._send_versioned_transaction(ixs)

    async def place_orders_for_market(
        self,
        asset: Asset,
        orders: list[OrderArgs],
        pre_instructions: Optional[list[Instruction]] = None,
        post_instructions: Optional[list[Instruction]] = None,
        tif_buffer: int = 0,
        priority_fee: int = 0,
    ):
        """
        Place orders for a market.

        Args:
            asset (Asset): The asset for which to place the orders.
            orders (list[OrderArgs]): The list of orders to place.
            pre_instructions (Optional[list[Instruction]], optional): The list of instructions to execute before
                placing the orders. Defaults to None.
            post_instructions (Optional[list[Instruction]], optional): The list of instructions to execute after
                placing the orders. Defaults to None.
            tif_buffer (int): Extra value to add to tif_expiry at epoch rollover to aid a smooth transition.
                Defaults to 0.
            priority_fee (int): Additional priority fee, in microlamports per CU. Defaults to 0.

        Returns:
            Transaction: The transaction of the placed orders.
        """
        # TODO: warn about log truncation above 10 orders
        ixs = []
        if not await self._check_open_orders_account_exists(asset):
            self._logger.info("User has no open orders account, creating one...")
            ixs.append(self._init_open_orders_ix(asset))

        if priority_fee > 0:
            ixs.append(set_compute_unit_price(priority_fee))

        if pre_instructions is not None:
            ixs.extend(pre_instructions)
        for order in orders:
            ixs.append(self._place_order_ix(asset, order.price, order.size, order.side, order.order_opts, tif_buffer))
        if post_instructions is not None:
            ixs.extend(post_instructions)
        self._logger.info(f"Placing {len(orders)} orders for {asset}")
        return await self._send_versioned_transaction(ixs)

    async def place_multi_orders_for_market(
        self,
        asset: Asset,
        bid_orders: list[MultiOrderArgs],
        ask_orders: list[MultiOrderArgs],
        order_type: OrderType,
        pre_instructions: Optional[list[Instruction]] = None,
        post_instructions: Optional[list[Instruction]] = None,
        tif_buffer: int = 0,
        priority_fee: int = 0,
    ):
        """
        Place multi orders for a market.

        Args:
            asset (Asset): The asset for which to place the orders.
            orders (list[OrderArgs]): The list of orders to place.
            pre_instructions (Optional[list[Instruction]], optional): The list of instructions to execute before
                placing the orders. Defaults to None.
            post_instructions (Optional[list[Instruction]], optional): The list of instructions to execute after
                placing the orders. Defaults to None.
            tif_buffer (int): Extra value to add to tif_expiry at epoch rollover to aid a smooth transition.
                Defaults to 0.
            priority_fee (int): Additional priority fee, in microlamports per CU. Defaults to 0.

        Returns:
            Transaction: The transaction of the placed orders.
        """
        # TODO: warn about log truncation above 10 orders
        ixs = []
        if not await self._check_open_orders_account_exists(asset):
            self._logger.info("User has no open orders account, creating one...")
            ixs.append(self._init_open_orders_ix(asset))

        if priority_fee > 0:
            ixs.append(set_compute_unit_price(priority_fee))

        if pre_instructions is not None:
            ixs.extend(pre_instructions)

        ixs.append(self._place_multi_orders_ix(asset, bid_orders, ask_orders, order_type, tif_buffer))

        if post_instructions is not None:
            ixs.extend(post_instructions)
        self._logger.info(f"Placing {len(bid_orders) + len(ask_orders)} orders using place_multi_orders for {asset}")
        return await self._send_versioned_transaction(ixs)

    async def replace_orders_for_market(self, asset: Asset, orders: list[OrderArgs], priority_fee: int = 0):
        """
        Replace orders for a market (atomically cancel all orders and replace them).

        Args:
            asset (Asset): The asset for which to replace the orders.
            orders (list[OrderArgs]): The list of new orders to place.
            priority_fee (int): Additional priority fee, in microlamports per CU. Defaults to 0.

        Returns:
            Transaction: The transaction of the replaced orders.
        """
        pre_ixs = []
        if priority_fee > 0:
            pre_ixs.append(set_compute_unit_price(priority_fee))
        pre_ixs.extend([self._cancel_orders_for_market_ix(asset)])
        return await self.place_orders_for_market(asset, orders, pre_instructions=pre_ixs)

    # TODO: liquidate
    async def liquidate(self):
        """
        Liquidate method.

        Raises:
            NotImplementedError: This method is not implemented yet.
        """
        raise NotImplementedError

    async def send_jito_tx(self, tx: VersionedTransaction, blockhash):
        client = await get_async_searcher_client("mainnet.block-engine.jito.wtf", self.provider.wallet.payer)

        # The tx to Jito and to RPC needs to be the same exact tx, so we simply add a 2nd tx for the Jito tip to the bundle
        # Note: Jito ignores any compute unit budget (prio fee) instructions, so we keep them in there
        msg = MessageV0.try_compile(
            self.provider.wallet.public_key,
            [
                transfer(
                    TransferParams(
                        from_pubkey=self.provider.wallet.public_key,
                        to_pubkey=Pubkey.from_string("DttWaMuVvTiduZRnguLF7jNxTgiMBZ1hyAumKUiL2KRL"),
                        lamports=self.jito_tip,
                    )
                )
            ],
            [constants.ZETA_LUT[self.network]],
            blockhash,
        )
        tx_tip = VersionedTransaction(msg, [self.provider.wallet.payer])

        try:
            jito_response = await client.SendBundle(
                SendBundleRequest(
                    bundle=Bundle(
                        header=None,
                        packets=[
                            Packet(
                                data=bytes(tx),
                                meta=Meta(size=len(bytes(tx)), addr="0.0.0.0", port=0, flags=None, sender_stake=0),
                            ),
                            Packet(
                                data=bytes(tx_tip),
                                meta=Meta(size=len(bytes(tx_tip)), addr="0.0.0.0", port=0, flags=None, sender_stake=0),
                            ),
                        ],
                    )
                )
            )
            self._logger.debug(f"Jito response: {jito_response}")
        except Exception as e:
            print(f"Jito error: {e}")

    async def _send_versioned_transaction(self, ixs: list[Instruction]):
        """
        Send a versioned transaction.

        Args:
            ixs (list[Instruction]): The list of instructions to include in the transaction.

        Returns:
            str: The signature(s) of the transaction(s).
        """
        # Prefetch blockhash, using cache if available
        if self.blockhash_cache:
            try:
                recent_blockhash = self.blockhash_cache.get()
                last_valid_block_height = None
                self._logger.debug(f"Blockhash cache hit, using cached blockhash: {recent_blockhash}")
            except ValueError:
                blockhash_resp = await self.connection.get_latest_blockhash(self.connection.commitment)
                recent_blockhash = self.connection.parse_recent_blockhash(blockhash_resp)
                last_valid_block_height = blockhash_resp.value.last_valid_block_height
                self._logger.debug(f"Blockhash cache miss, fetched from RPC: {recent_blockhash}")
        else:
            blockhash_resp = await self.connection.get_latest_blockhash(self.connection.commitment)
            recent_blockhash = self.connection.parse_recent_blockhash(blockhash_resp)
            last_valid_block_height = blockhash_resp.value.last_valid_block_height
            self._logger.debug(f"Blockhash cache not enabled, fetched from RPC: {recent_blockhash}")

        msg = MessageV0.try_compile(
            self.provider.wallet.public_key, ixs, [constants.ZETA_LUT[self.network]], recent_blockhash
        )
        tx = VersionedTransaction(msg, [self.provider.wallet.payer])

        try:
            opts = self.provider.opts._replace(last_valid_block_height=last_valid_block_height)
            if len(self.double_down_providers) > 0:
                tasks = []
                if self.double_down_jito:
                    tasks.append(self.send_jito_tx(tx, recent_blockhash))
                for provider in self.double_down_providers:
                    tasks.append(provider.send(tx, opts))
                signature = await asyncio.gather(*tasks)
            else:
                signature = [await self.provider.send(tx, opts)]
        except RPCException as exc:
            # This won't work on zDEX errors
            # TODO: add ZDEX error parsing
            parsed = from_tx_error(exc)
            self._logger.error(parsed)
            if parsed is not None:
                raise parsed from exc
            raise exc
        return signature



from __future__ import annotations

import typing

from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey

from ..program_id import PROGRAM_ID


class CloseReferrerAccountsAccounts(typing.TypedDict):
    referrer_id_account: Pubkey
    referrer_pubkey_account: Pubkey
    authority: Pubkey


def close_referrer_accounts(
    accounts: CloseReferrerAccountsAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["referrer_id_account"], is_signer=False, is_writable=True),
        AccountMeta(
            pubkey=accounts["referrer_pubkey_account"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(pubkey=accounts["authority"], is_signer=True, is_writable=True),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xe0N7\x8b\xcb\xec>N"
    encoded_args = b""
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations

import typing

import borsh_construct as borsh
from construct import Construct
from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey

from .. import types
from ..program_id import PROGRAM_ID


class PositionMovementArgs(typing.TypedDict):
    movement_type: types.movement_type.MovementTypeKind
    movements: list[types.position_movement_arg.PositionMovementArg]


layout = borsh.CStruct(
    "movement_type" / types.movement_type.layout,
    "movements" / borsh.Vec(typing.cast(Construct, types.position_movement_arg.PositionMovementArg.layout)),
)


class PositionMovementAccounts(typing.TypedDict):
    state: Pubkey
    zeta_group: Pubkey
    margin_account: Pubkey
    spread_account: Pubkey
    authority: Pubkey
    greeks: Pubkey
    oracle: Pubkey
    oracle_backup_feed: Pubkey
    oracle_backup_program: Pubkey


def position_movement(
    args: PositionMovementArgs,
    accounts: PositionMovementAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["state"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["zeta_group"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["margin_account"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["spread_account"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["authority"], is_signer=True, is_writable=False),
        AccountMeta(pubkey=accounts["greeks"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["oracle"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["oracle_backup_feed"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["oracle_backup_program"], is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"u\x10K\xf9\xb3\x7f\xab\x93"
    encoded_args = layout.build(
        {
            "movement_type": args["movement_type"].to_encodable(),
            "movements": list(map(lambda item: item.to_encodable(), args["movements"])),
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations

import typing

import borsh_construct as borsh
from anchorpy.borsh_extension import BorshPubkey
from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey

from ..program_id import PROGRAM_ID


class EditDelegatedPubkeyArgs(typing.TypedDict):
    new_key: Pubkey


layout = borsh.CStruct("new_key" / BorshPubkey)


class EditDelegatedPubkeyAccounts(typing.TypedDict):
    margin_account: Pubkey
    authority: Pubkey


def edit_delegated_pubkey(
    args: EditDelegatedPubkeyArgs,
    accounts: EditDelegatedPubkeyAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["margin_account"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["authority"], is_signer=True, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\x89\xf5GY.\xf9\x165"
    encoded_args = layout.build(
        {
            "new_key": args["new_key"],
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations

import typing

import borsh_construct as borsh
from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey

from ..program_id import PROGRAM_ID


class TakeTriggerOrderArgs(typing.TypedDict):
    trigger_order_bit: int


layout = borsh.CStruct("trigger_order_bit" / borsh.U8)


class TakeTriggerOrderAccounts(typing.TypedDict):
    trigger_order: Pubkey
    state: Pubkey
    pricing: Pubkey
    oracle: Pubkey
    oracle_backup_feed: Pubkey
    oracle_backup_program: Pubkey
    bids: Pubkey
    asks: Pubkey
    taker: Pubkey
    taker_margin_account: Pubkey
    order_margin_account: Pubkey


def take_trigger_order(
    args: TakeTriggerOrderArgs,
    accounts: TakeTriggerOrderAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["trigger_order"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["state"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["pricing"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["oracle"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["oracle_backup_feed"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["oracle_backup_program"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["bids"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["asks"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["taker"], is_signer=True, is_writable=True),
        AccountMeta(pubkey=accounts["taker_margin_account"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["order_margin_account"], is_signer=False, is_writable=True),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"k\xcf;\xe2\x19\x17\x1f\xa1"
    encoded_args = layout.build(
        {
            "trigger_order_bit": args["trigger_order_bit"],
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations

import typing

import borsh_construct as borsh
from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey

from .. import types
from ..program_id import PROGRAM_ID


class UpdatePricingV2Args(typing.TypedDict):
    asset: types.asset.AssetKind


layout = borsh.CStruct("asset" / types.asset.layout)


class UpdatePricingV2Accounts(typing.TypedDict):
    state: Pubkey
    pricing: Pubkey
    oracle: Pubkey
    oracle_backup_feed: Pubkey
    oracle_backup_program: Pubkey
    perp_market: Pubkey
    perp_bids: Pubkey
    perp_asks: Pubkey


def update_pricing_v2(
    args: UpdatePricingV2Args,
    accounts: UpdatePricingV2Accounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["state"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["pricing"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["oracle"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["oracle_backup_feed"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["oracle_backup_program"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["perp_market"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["perp_bids"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["perp_asks"], is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xebm\x8a\xad\x0f%3\xf4"
    encoded_args = layout.build(
        {
            "asset": args["asset"].to_encodable(),
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from anchorpy.borsh_extension import BorshPubkey
from construct import Container
from solders.pubkey import Pubkey


class SetReferralsRewardsArgsJSON(typing.TypedDict):
    referrals_account_key: str
    pending_rewards: int
    overwrite: bool


@dataclass
class SetReferralsRewardsArgs:
    layout: typing.ClassVar = borsh.CStruct(
        "referrals_account_key" / BorshPubkey,
        "pending_rewards" / borsh.U64,
        "overwrite" / borsh.Bool,
    )
    referrals_account_key: Pubkey
    pending_rewards: int
    overwrite: bool

    @classmethod
    def from_decoded(cls, obj: Container) -> "SetReferralsRewardsArgs":
        return cls(
            referrals_account_key=obj.referrals_account_key,
            pending_rewards=obj.pending_rewards,
            overwrite=obj.overwrite,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "referrals_account_key": self.referrals_account_key,
            "pending_rewards": self.pending_rewards,
            "overwrite": self.overwrite,
        }

    def to_json(self) -> SetReferralsRewardsArgsJSON:
        return {
            "referrals_account_key": str(self.referrals_account_key),
            "pending_rewards": self.pending_rewards,
            "overwrite": self.overwrite,
        }

    @classmethod
    def from_json(cls, obj: SetReferralsRewardsArgsJSON) -> "SetReferralsRewardsArgs":
        return cls(
            referrals_account_key=Pubkey.from_string(obj["referrals_account_key"]),
            pending_rewards=obj["pending_rewards"],
            overwrite=obj["overwrite"],
        )



from __future__ import annotations

import typing

import borsh_construct as borsh
from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey
from solders.sysvar import RENT
from spl.token.constants import TOKEN_PROGRAM_ID

from ..program_id import PROGRAM_ID


class ExecuteTriggerOrderArgs(typing.TypedDict):
    trigger_order_bit: int


layout = borsh.CStruct("trigger_order_bit" / borsh.U8)


class ExecuteTriggerOrderAccounts(typing.TypedDict):
    admin: Pubkey
    trigger_order: Pubkey
    place_order_accounts: PlaceOrderAccountsNested


class PlaceOrderAccountsNested(typing.TypedDict):
    state: Pubkey
    pricing: Pubkey
    margin_account: Pubkey
    dex_program: Pubkey
    serum_authority: Pubkey
    open_orders: Pubkey
    market_accounts: MarketAccountsNested
    oracle: Pubkey
    oracle_backup_feed: Pubkey
    oracle_backup_program: Pubkey
    market_mint: Pubkey
    mint_authority: Pubkey
    perp_sync_queue: Pubkey


class MarketAccountsNested(typing.TypedDict):
    market: Pubkey
    request_queue: Pubkey
    event_queue: Pubkey
    bids: Pubkey
    asks: Pubkey
    order_payer_token_account: Pubkey
    coin_vault: Pubkey
    pc_vault: Pubkey
    coin_wallet: Pubkey
    pc_wallet: Pubkey


def execute_trigger_order(
    args: ExecuteTriggerOrderArgs,
    accounts: ExecuteTriggerOrderAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["admin"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["trigger_order"], is_signer=False, is_writable=True),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["state"],
            is_signer=False,
            is_writable=False,
        ),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["pricing"],
            is_signer=False,
            is_writable=False,
        ),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["margin_account"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["dex_program"],
            is_signer=False,
            is_writable=False,
        ),
        AccountMeta(pubkey=TOKEN_PROGRAM_ID, is_signer=False, is_writable=False),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["serum_authority"],
            is_signer=False,
            is_writable=False,
        ),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["open_orders"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(pubkey=RENT, is_signer=False, is_writable=False),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["market_accounts"]["market"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["market_accounts"]["request_queue"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["market_accounts"]["event_queue"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["market_accounts"]["bids"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["market_accounts"]["asks"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["market_accounts"]["order_payer_token_account"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["market_accounts"]["coin_vault"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["market_accounts"]["pc_vault"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["market_accounts"]["coin_wallet"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["market_accounts"]["pc_wallet"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["oracle"],
            is_signer=False,
            is_writable=False,
        ),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["oracle_backup_feed"],
            is_signer=False,
            is_writable=False,
        ),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["oracle_backup_program"],
            is_signer=False,
            is_writable=False,
        ),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["market_mint"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["mint_authority"],
            is_signer=False,
            is_writable=False,
        ),
        AccountMeta(
            pubkey=accounts["place_order_accounts"]["perp_sync_queue"],
            is_signer=False,
            is_writable=True,
        ),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"i\nh\x88\xd7\x86T\xab"
    encoded_args = layout.build(
        {
            "trigger_order_bit": args["trigger_order_bit"],
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from construct import Container


class InitializeZetaPricingArgsJSON(typing.TypedDict):
    min_funding_rate_percent: int
    max_funding_rate_percent: int
    perp_impact_cash_delta: int
    margin_initial: int
    margin_maintenance: int
    pricing_nonce: int


@dataclass
class InitializeZetaPricingArgs:
    layout: typing.ClassVar = borsh.CStruct(
        "min_funding_rate_percent" / borsh.I64,
        "max_funding_rate_percent" / borsh.I64,
        "perp_impact_cash_delta" / borsh.U64,
        "margin_initial" / borsh.U64,
        "margin_maintenance" / borsh.U64,
        "pricing_nonce" / borsh.U8,
    )
    min_funding_rate_percent: int
    max_funding_rate_percent: int
    perp_impact_cash_delta: int
    margin_initial: int
    margin_maintenance: int
    pricing_nonce: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "InitializeZetaPricingArgs":
        return cls(
            min_funding_rate_percent=obj.min_funding_rate_percent,
            max_funding_rate_percent=obj.max_funding_rate_percent,
            perp_impact_cash_delta=obj.perp_impact_cash_delta,
            margin_initial=obj.margin_initial,
            margin_maintenance=obj.margin_maintenance,
            pricing_nonce=obj.pricing_nonce,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "min_funding_rate_percent": self.min_funding_rate_percent,
            "max_funding_rate_percent": self.max_funding_rate_percent,
            "perp_impact_cash_delta": self.perp_impact_cash_delta,
            "margin_initial": self.margin_initial,
            "margin_maintenance": self.margin_maintenance,
            "pricing_nonce": self.pricing_nonce,
        }

    def to_json(self) -> InitializeZetaPricingArgsJSON:
        return {
            "min_funding_rate_percent": self.min_funding_rate_percent,
            "max_funding_rate_percent": self.max_funding_rate_percent,
            "perp_impact_cash_delta": self.perp_impact_cash_delta,
            "margin_initial": self.margin_initial,
            "margin_maintenance": self.margin_maintenance,
            "pricing_nonce": self.pricing_nonce,
        }

    @classmethod
    def from_json(cls, obj: InitializeZetaPricingArgsJSON) -> "InitializeZetaPricingArgs":
        return cls(
            min_funding_rate_percent=obj["min_funding_rate_percent"],
            max_funding_rate_percent=obj["max_funding_rate_percent"],
            perp_impact_cash_delta=obj["perp_impact_cash_delta"],
            margin_initial=obj["margin_initial"],
            margin_maintenance=obj["margin_maintenance"],
            pricing_nonce=obj["pricing_nonce"],
        )



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from construct import Container


class StrikeJSON(typing.TypedDict):
    is_set: bool
    value: int


@dataclass
class Strike:
    layout: typing.ClassVar = borsh.CStruct("is_set" / borsh.Bool, "value" / borsh.U64)
    is_set: bool
    value: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "Strike":
        return cls(is_set=obj.is_set, value=obj.value)

    def to_encodable(self) -> dict[str, typing.Any]:
        return {"is_set": self.is_set, "value": self.value}

    def to_json(self) -> StrikeJSON:
        return {"is_set": self.is_set, "value": self.value}

    @classmethod
    def from_json(cls, obj: StrikeJSON) -> "Strike":
        return cls(is_set=obj["is_set"], value=obj["value"])



from .cross_margin_account import CrossMarginAccount, CrossMarginAccountJSON
from .cross_margin_account_manager import (
    CrossMarginAccountManager,
    CrossMarginAccountManagerJSON,
)
from .cross_open_orders_map import CrossOpenOrdersMap, CrossOpenOrdersMapJSON
from .greeks import Greeks, GreeksJSON
from .insurance_deposit_account import (
    InsuranceDepositAccount,
    InsuranceDepositAccountJSON,
)
from .margin_account import MarginAccount, MarginAccountJSON
from .market_indexes import MarketIndexes, MarketIndexesJSON
from .market_node import MarketNode, MarketNodeJSON
from .open_orders_map import OpenOrdersMap, OpenOrdersMapJSON
from .perp_sync_queue import PerpSyncQueue, PerpSyncQueueJSON
from .pricing import Pricing, PricingJSON
from .referrer_id_account import ReferrerIdAccount, ReferrerIdAccountJSON
from .referrer_pubkey_account import ReferrerPubkeyAccount, ReferrerPubkeyAccountJSON
from .settlement_account import SettlementAccount, SettlementAccountJSON
from .socialized_loss_account import SocializedLossAccount, SocializedLossAccountJSON
from .spread_account import SpreadAccount, SpreadAccountJSON
from .state import State, StateJSON
from .trigger_order import TriggerOrder, TriggerOrderJSON
from .underlying import Underlying, UnderlyingJSON
from .whitelist_deposit_account import (
    WhitelistDepositAccount,
    WhitelistDepositAccountJSON,
)
from .whitelist_insurance_account import (
    WhitelistInsuranceAccount,
    WhitelistInsuranceAccountJSON,
)
from .whitelist_trading_fees_account import (
    WhitelistTradingFeesAccount,
    WhitelistTradingFeesAccountJSON,
)
from .zeta_group import ZetaGroup, ZetaGroupJSON



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from anchorpy.borsh_extension import BorshPubkey
from construct import Container
from solders.pubkey import Pubkey

from . import kind, strike


class ProductJSON(typing.TypedDict):
    market: str
    strike: strike.StrikeJSON
    dirty: bool
    kind: kind.KindJSON


@dataclass
class Product:
    layout: typing.ClassVar = borsh.CStruct(
        "market" / BorshPubkey,
        "strike" / strike.Strike.layout,
        "dirty" / borsh.Bool,
        "kind" / kind.layout,
    )
    market: Pubkey
    strike: strike.Strike
    dirty: bool
    kind: kind.KindKind

    @classmethod
    def from_decoded(cls, obj: Container) -> "Product":
        return cls(
            market=obj.market,
            strike=strike.Strike.from_decoded(obj.strike),
            dirty=obj.dirty,
            kind=kind.from_decoded(obj.kind),
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "market": self.market,
            "strike": self.strike.to_encodable(),
            "dirty": self.dirty,
            "kind": self.kind.to_encodable(),
        }

    def to_json(self) -> ProductJSON:
        return {
            "market": str(self.market),
            "strike": self.strike.to_json(),
            "dirty": self.dirty,
            "kind": self.kind.to_json(),
        }

    @classmethod
    def from_json(cls, obj: ProductJSON) -> "Product":
        return cls(
            market=Pubkey.from_string(obj["market"]),
            strike=strike.Strike.from_json(obj["strike"]),
            dirty=obj["dirty"],
            kind=kind.from_json(obj["kind"]),
        )



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from construct import Container


class UpdateVolatilityArgsJSON(typing.TypedDict):
    expiry_index: int
    volatility: list[int]


@dataclass
class UpdateVolatilityArgs:
    layout: typing.ClassVar = borsh.CStruct("expiry_index" / borsh.U8, "volatility" / borsh.U64[5])
    expiry_index: int
    volatility: list[int]

    @classmethod
    def from_decoded(cls, obj: Container) -> "UpdateVolatilityArgs":
        return cls(expiry_index=obj.expiry_index, volatility=obj.volatility)

    def to_encodable(self) -> dict[str, typing.Any]:
        return {"expiry_index": self.expiry_index, "volatility": self.volatility}

    def to_json(self) -> UpdateVolatilityArgsJSON:
        return {"expiry_index": self.expiry_index, "volatility": self.volatility}

    @classmethod
    def from_json(cls, obj: UpdateVolatilityArgsJSON) -> "UpdateVolatilityArgs":
        return cls(expiry_index=obj["expiry_index"], volatility=obj["volatility"])



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from anchorpy.borsh_extension import EnumForCodegen


class CancelJSON(typing.TypedDict):
    kind: typing.Literal["Cancel"]


class FillJSON(typing.TypedDict):
    kind: typing.Literal["Fill"]


class BootedJSON(typing.TypedDict):
    kind: typing.Literal["Booted"]


@dataclass
class Cancel:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "Cancel"

    @classmethod
    def to_json(cls) -> CancelJSON:
        return CancelJSON(
            kind="Cancel",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Cancel": {},
        }


@dataclass
class Fill:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "Fill"

    @classmethod
    def to_json(cls) -> FillJSON:
        return FillJSON(
            kind="Fill",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Fill": {},
        }


@dataclass
class Booted:
    discriminator: typing.ClassVar = 2
    kind: typing.ClassVar = "Booted"

    @classmethod
    def to_json(cls) -> BootedJSON:
        return BootedJSON(
            kind="Booted",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Booted": {},
        }


OrderCompleteTypeKind = typing.Union[Cancel, Fill, Booted]
OrderCompleteTypeJSON = typing.Union[CancelJSON, FillJSON, BootedJSON]


def from_decoded(obj: dict) -> OrderCompleteTypeKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "Cancel" in obj:
        return Cancel()
    if "Fill" in obj:
        return Fill()
    if "Booted" in obj:
        return Booted()
    raise ValueError("Invalid enum object")


def from_json(obj: OrderCompleteTypeJSON) -> OrderCompleteTypeKind:
    if obj["kind"] == "Cancel":
        return Cancel()
    if obj["kind"] == "Fill":
        return Fill()
    if obj["kind"] == "Booted":
        return Booted()
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen("Cancel" / borsh.CStruct(), "Fill" / borsh.CStruct(), "Booted" / borsh.CStruct())



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from construct import Container


class HaltStateJSON(typing.TypedDict):
    halted: bool
    spot_price: int
    timestamp: int
    mark_prices_set: list[bool]
    mark_prices_set_padding: list[bool]
    perp_mark_price_set: bool
    market_nodes_cleaned: list[bool]
    market_nodes_cleaned_padding: list[bool]
    market_cleaned: list[bool]
    market_cleaned_padding: list[bool]
    perp_market_cleaned: bool


@dataclass
class HaltState:
    layout: typing.ClassVar = borsh.CStruct(
        "halted" / borsh.Bool,
        "spot_price" / borsh.U64,
        "timestamp" / borsh.U64,
        "mark_prices_set" / borsh.Bool[2],
        "mark_prices_set_padding" / borsh.Bool[3],
        "perp_mark_price_set" / borsh.Bool,
        "market_nodes_cleaned" / borsh.Bool[2],
        "market_nodes_cleaned_padding" / borsh.Bool[4],
        "market_cleaned" / borsh.Bool[46],
        "market_cleaned_padding" / borsh.Bool[91],
        "perp_market_cleaned" / borsh.Bool,
    )
    halted: bool
    spot_price: int
    timestamp: int
    mark_prices_set: list[bool]
    mark_prices_set_padding: list[bool]
    perp_mark_price_set: bool
    market_nodes_cleaned: list[bool]
    market_nodes_cleaned_padding: list[bool]
    market_cleaned: list[bool]
    market_cleaned_padding: list[bool]
    perp_market_cleaned: bool

    @classmethod
    def from_decoded(cls, obj: Container) -> "HaltState":
        return cls(
            halted=obj.halted,
            spot_price=obj.spot_price,
            timestamp=obj.timestamp,
            mark_prices_set=obj.mark_prices_set,
            mark_prices_set_padding=obj.mark_prices_set_padding,
            perp_mark_price_set=obj.perp_mark_price_set,
            market_nodes_cleaned=obj.market_nodes_cleaned,
            market_nodes_cleaned_padding=obj.market_nodes_cleaned_padding,
            market_cleaned=obj.market_cleaned,
            market_cleaned_padding=obj.market_cleaned_padding,
            perp_market_cleaned=obj.perp_market_cleaned,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "halted": self.halted,
            "spot_price": self.spot_price,
            "timestamp": self.timestamp,
            "mark_prices_set": self.mark_prices_set,
            "mark_prices_set_padding": self.mark_prices_set_padding,
            "perp_mark_price_set": self.perp_mark_price_set,
            "market_nodes_cleaned": self.market_nodes_cleaned,
            "market_nodes_cleaned_padding": self.market_nodes_cleaned_padding,
            "market_cleaned": self.market_cleaned,
            "market_cleaned_padding": self.market_cleaned_padding,
            "perp_market_cleaned": self.perp_market_cleaned,
        }

    def to_json(self) -> HaltStateJSON:
        return {
            "halted": self.halted,
            "spot_price": self.spot_price,
            "timestamp": self.timestamp,
            "mark_prices_set": self.mark_prices_set,
            "mark_prices_set_padding": self.mark_prices_set_padding,
            "perp_mark_price_set": self.perp_mark_price_set,
            "market_nodes_cleaned": self.market_nodes_cleaned,
            "market_nodes_cleaned_padding": self.market_nodes_cleaned_padding,
            "market_cleaned": self.market_cleaned,
            "market_cleaned_padding": self.market_cleaned_padding,
            "perp_market_cleaned": self.perp_market_cleaned,
        }

    @classmethod
    def from_json(cls, obj: HaltStateJSON) -> "HaltState":
        return cls(
            halted=obj["halted"],
            spot_price=obj["spot_price"],
            timestamp=obj["timestamp"],
            mark_prices_set=obj["mark_prices_set"],
            mark_prices_set_padding=obj["mark_prices_set_padding"],
            perp_mark_price_set=obj["perp_mark_price_set"],
            market_nodes_cleaned=obj["market_nodes_cleaned"],
            market_nodes_cleaned_padding=obj["market_nodes_cleaned_padding"],
            market_cleaned=obj["market_cleaned"],
            market_cleaned_padding=obj["market_cleaned_padding"],
            perp_market_cleaned=obj["perp_market_cleaned"],
        )



import time
from typing import Iterable, Optional, Union

from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
from solders.pubkey import Pubkey

from zetamarkets_py import constants, utils
from zetamarkets_py.serum_client.accounts.market_state import MarketState
from zetamarkets_py.serum_client.accounts.orderbook import OrderbookAccount
from zetamarkets_py.serum_client.types.slab import SlabInnerNode, SlabLeafNode
from zetamarkets_py.types import Network, Order, OrderInfo, Side


class Orderbook:
    """Represents an order book.

    Attributes:
        side (Side): The side of the order book.
    """

    def __init__(self, side: Side, orderbook: OrderbookAccount, market_state: MarketState) -> None:
        """Initializes the Orderbook class.

        Args:
            side (Side): The side of the order book.
            orderbook (OrderbookAccount): The orderbook account.
            market_state (MarketState): The market state.

        Raises:
            Exception: If the order book is not initialized or neither of bids or asks.
        """
        if not (orderbook.account_flags.initialized and orderbook.account_flags.bids ^ orderbook.account_flags.asks):
            raise Exception("Invalid order book, either not initialized or neither of bids or asks")
        self.side = side
        self._slab = orderbook.slab
        self._market_state = market_state

    @classmethod
    async def load(
        cls,
        conn: AsyncClient,
        address: Pubkey,
        commitment: Commitment,
        side: Side,
        market_state: MarketState,
        program_id: Pubkey = constants.MATCHING_ENGINE_PID[Network.MAINNET],
    ) -> Optional["Orderbook"]:
        """Loads the order book.

        Args:
            conn (AsyncClient): The connection.
            address (Pubkey): The address.
            commitment (Commitment): The commitment.
            side (Side): The side.
            market_state (MarketState): The market state.
            program_id (Pubkey, optional): The program ID. Defaults to constants.MATCHING_ENGINE_PID[Network.MAINNET].

        Returns:
            Optional[Orderbook]: The order book if it exists, None otherwise.
        """
        orderbook = await OrderbookAccount.fetch(conn, address, commitment, program_id)
        if orderbook is None:
            return None
        return cls(side, orderbook, market_state)

    @staticmethod
    def _get_price_from_slab(node: Union[SlabInnerNode, SlabLeafNode]) -> int:
        """Gets the price from a slab node key.

        Args:
            node (Union[SlabInnerNode, SlabLeafNode]): The slab node.

        Returns:
            int: The price.
        """
        return node.key >> 64

    @staticmethod
    def _get_seq_num_from_slab(key: int, side: Side) -> int:
        """Gets the sequence number from a slab node key.

        Args:
            key (int): The key.
            side (Side): The side.

        Returns:
            int: The sequence number.
        """
        # TODO: bit length might be variable bc of python
        # if key.bit_length() != 128:
        #     print(key.bit_length())
        #     raise ValueError("Key should be 128 bits")
        # Mask off highest 64 bits
        UPPER_64_BITMASK = 0x0000000000000000FFFFFFFFFFFFFFFF
        lower = key & UPPER_64_BITMASK
        if side == Side.Bid:
            # Bitwise NOT (since python doesn't have unsigned ints)
            return (1 << 64) - 1 - lower
        else:
            return lower

    def _is_order_expired(
        self, clock_ts: int, tif_offset: int, epoch_length: int, seq_num: int, epoch_start_seq_num: int, tif_buffer: int
    ) -> int:
        """Checks if the order is expired.

        Args:
            clock_ts (int): The clock timestamp.
            tif_offset (int): The time in force offset.
            epoch_start_ts (int): The epoch start timestamp.
            seq_num (int): The sequence number.
            epoch_start_seq_num (int): The epoch start sequence number.

        Returns:
            int: 1 if the order is expired, 0 otherwise.
        """
        if tif_offset > 0:
            # Add TIF buffer here to get into the next epoch earlier
            epoch_start_ts = (clock_ts + tif_buffer) - (clock_ts + tif_buffer) % epoch_length

            # Add TIF buffer here to account for clock drift when not around the epoch crossover
            if epoch_start_ts + tif_offset + tif_buffer < clock_ts or seq_num <= epoch_start_seq_num:
                return True

        return False

    # using local time as a hack as opposed to self.exchange.clock.account.unix_timestamp
    def _get_l2(self, depth: int, clock_ts: int = int(time.time()), tif_buffer: int = 10) -> list[OrderInfo]:
        """Gets the Level 2 market information.

        Args:
            depth (int): The depth.
            clock_ts (int, optional): The clock timestamp. Defaults to int(time.time()).

        Returns:
            list[OrderInfo]: The Level 2 market information.
        """
        descending = self.side == Side.Bid
        # The first element of the inner list is price, the second is quantity.
        levels: list[list[int]] = []
        for node in self._slab.items(descending):
            seq_num = self._get_seq_num_from_slab(node.key, self.side)
            order_expired = self._is_order_expired(
                clock_ts,
                node.tif_offset,
                self._market_state.epoch_length,
                seq_num,
                self._market_state.start_epoch_seq_num,
                tif_buffer,
            )
            if order_expired:
                continue
            price = self._get_price_from_slab(node)
            if len(levels) > 0 and levels[len(levels) - 1][0] == price:
                levels[len(levels) - 1][1] += node.quantity
            elif len(levels) == depth:
                break
            else:
                levels.append([price, node.quantity])
        return [
            OrderInfo(
                price=utils.convert_fixed_int_to_decimal(price_lots),
                size=utils.convert_fixed_lot_to_decimal(size_lots),
            )
            for price_lots, size_lots in levels
        ]

    def __iter__(self) -> Iterable[Order]:
        """Returns an iterator over the orders.

        Returns:
            Iterable[Order]: An iterator over the orders.
        """
        return self.orders()

    def orders(self) -> Iterable[Order]:
        """Yields the orders.

        Yields:
            Order: The order.
        """
        for node in self._slab.items():
            price = self._get_price_from_slab(node)
            open_orders_address = node.owner

            yield Order(
                order_id=node.key,
                client_id=node.client_order_id,
                open_order_address=open_orders_address,
                fee_tier=node.fee_tier,
                info=OrderInfo(
                    price=utils.convert_fixed_int_to_decimal(price),
                    size=utils.convert_fixed_lot_to_decimal(node.quantity),
                ),
                side=self.side,
                open_order_slot=node.owner_slot,
                tif_offset=node.tif_offset,
            )



from __future__ import annotations

import typing

import borsh_construct as borsh
from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey

from ..program_id import PROGRAM_ID


class ToggleMarketMakerArgs(typing.TypedDict):
    is_market_maker: bool


layout = borsh.CStruct("is_market_maker" / borsh.Bool)


class ToggleMarketMakerAccounts(typing.TypedDict):
    state: Pubkey
    admin: Pubkey
    margin_account: Pubkey


def toggle_market_maker(
    args: ToggleMarketMakerArgs,
    accounts: ToggleMarketMakerAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["state"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["admin"], is_signer=True, is_writable=False),
        AccountMeta(pubkey=accounts["margin_account"], is_signer=False, is_writable=True),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xcb\xf7T\x9fh\xfd\x94P"
    encoded_args = layout.build(
        {
            "is_market_maker": args["is_market_maker"],
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations

import typing

import borsh_construct as borsh
from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey

from ..program_id import PROGRAM_ID


class LiquidateArgs(typing.TypedDict):
    size: int


layout = borsh.CStruct("size" / borsh.U64)


class LiquidateAccounts(typing.TypedDict):
    state: Pubkey
    liquidator: Pubkey
    liquidator_margin_account: Pubkey
    greeks: Pubkey
    oracle: Pubkey
    oracle_backup_feed: Pubkey
    oracle_backup_program: Pubkey
    market: Pubkey
    zeta_group: Pubkey
    liquidated_margin_account: Pubkey


def liquidate(
    args: LiquidateArgs,
    accounts: LiquidateAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["state"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["liquidator"], is_signer=True, is_writable=False),
        AccountMeta(
            pubkey=accounts["liquidator_margin_account"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(pubkey=accounts["greeks"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["oracle"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["oracle_backup_feed"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["oracle_backup_program"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["market"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["zeta_group"], is_signer=False, is_writable=False),
        AccountMeta(
            pubkey=accounts["liquidated_margin_account"],
            is_signer=False,
            is_writable=True,
        ),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xdf\xb3\xe2}0.'J"
    encoded_args = layout.build(
        {
            "size": args["size"],
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



import typing
from dataclasses import dataclass

import borsh_construct as borsh
from anchorpy.borsh_extension import BorshPubkey
from anchorpy.coder.accounts import ACCOUNT_DISCRIMINATOR_SIZE
from anchorpy.error import AccountInvalidDiscriminator
from anchorpy.utils.rpc import get_multiple_accounts
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
from solders.pubkey import Pubkey

from .. import types
from ..program_id import PROGRAM_ID


class MarginAccountJSON(typing.TypedDict):
    authority: str
    nonce: int
    balance: int
    force_cancel_flag: bool
    open_orders_nonce: list[int]
    series_expiry: list[int]
    series_expiry_padding: int
    product_ledgers: list[types.product_ledger.ProductLedgerJSON]
    product_ledgers_padding: list[types.product_ledger.ProductLedgerJSON]
    perp_product_ledger: types.product_ledger.ProductLedgerJSON
    rebalance_amount: int
    asset: types.asset.AssetJSON
    account_type: types.margin_account_type.MarginAccountTypeJSON
    last_funding_delta: types.anchor_decimal.AnchorDecimalJSON
    delegated_pubkey: str
    rebate_rebalance_amount: int
    padding: list[int]


@dataclass
class MarginAccount:
    discriminator: typing.ClassVar = b"\x85\xdc\xad\xd5\xb3\xd3+\xee"
    layout: typing.ClassVar = borsh.CStruct(
        "authority" / BorshPubkey,
        "nonce" / borsh.U8,
        "balance" / borsh.U64,
        "force_cancel_flag" / borsh.Bool,
        "open_orders_nonce" / borsh.U8[138],
        "series_expiry" / borsh.U64[5],
        "series_expiry_padding" / borsh.U64,
        "product_ledgers" / types.product_ledger.ProductLedger.layout[46],
        "product_ledgers_padding" / types.product_ledger.ProductLedger.layout[91],
        "perp_product_ledger" / types.product_ledger.ProductLedger.layout,
        "rebalance_amount" / borsh.I64,
        "asset" / types.asset.layout,
        "account_type" / types.margin_account_type.layout,
        "last_funding_delta" / types.anchor_decimal.AnchorDecimal.layout,
        "delegated_pubkey" / BorshPubkey,
        "rebate_rebalance_amount" / borsh.U64,
        "padding" / borsh.U8[330],
    )
    authority: Pubkey
    nonce: int
    balance: int
    force_cancel_flag: bool
    open_orders_nonce: list[int]
    series_expiry: list[int]
    series_expiry_padding: int
    product_ledgers: list[types.product_ledger.ProductLedger]
    product_ledgers_padding: list[types.product_ledger.ProductLedger]
    perp_product_ledger: types.product_ledger.ProductLedger
    rebalance_amount: int
    asset: types.asset.AssetKind
    account_type: types.margin_account_type.MarginAccountTypeKind
    last_funding_delta: types.anchor_decimal.AnchorDecimal
    delegated_pubkey: Pubkey
    rebate_rebalance_amount: int
    padding: list[int]

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        address: Pubkey,
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.Optional["MarginAccount"]:
        resp = await conn.get_account_info(address, commitment=commitment)
        info = resp.value
        if info is None:
            return None
        if info.owner != program_id:
            raise ValueError("Account does not belong to this program")
        bytes_data = info.data
        return cls.decode(bytes_data)

    @classmethod
    async def fetch_multiple(
        cls,
        conn: AsyncClient,
        addresses: list[Pubkey],
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.List[typing.Optional["MarginAccount"]]:
        infos = await get_multiple_accounts(conn, addresses, commitment=commitment)
        res: typing.List[typing.Optional["MarginAccount"]] = []
        for info in infos:
            if info is None:
                res.append(None)
                continue
            if info.account.owner != program_id:
                raise ValueError("Account does not belong to this program")
            res.append(cls.decode(info.account.data))
        return res

    @classmethod
    def decode(cls, data: bytes) -> "MarginAccount":
        if data[:ACCOUNT_DISCRIMINATOR_SIZE] != cls.discriminator:
            raise AccountInvalidDiscriminator("The discriminator for this account is invalid")
        dec = MarginAccount.layout.parse(data[ACCOUNT_DISCRIMINATOR_SIZE:])
        return cls(
            authority=dec.authority,
            nonce=dec.nonce,
            balance=dec.balance,
            force_cancel_flag=dec.force_cancel_flag,
            open_orders_nonce=dec.open_orders_nonce,
            series_expiry=dec.series_expiry,
            series_expiry_padding=dec.series_expiry_padding,
            product_ledgers=list(
                map(
                    lambda item: types.product_ledger.ProductLedger.from_decoded(item),
                    dec.product_ledgers,
                )
            ),
            product_ledgers_padding=list(
                map(
                    lambda item: types.product_ledger.ProductLedger.from_decoded(item),
                    dec.product_ledgers_padding,
                )
            ),
            perp_product_ledger=types.product_ledger.ProductLedger.from_decoded(dec.perp_product_ledger),
            rebalance_amount=dec.rebalance_amount,
            asset=types.asset.from_decoded(dec.asset),
            account_type=types.margin_account_type.from_decoded(dec.account_type),
            last_funding_delta=types.anchor_decimal.AnchorDecimal.from_decoded(dec.last_funding_delta),
            delegated_pubkey=dec.delegated_pubkey,
            rebate_rebalance_amount=dec.rebate_rebalance_amount,
            padding=dec.padding,
        )

    def to_json(self) -> MarginAccountJSON:
        return {
            "authority": str(self.authority),
            "nonce": self.nonce,
            "balance": self.balance,
            "force_cancel_flag": self.force_cancel_flag,
            "open_orders_nonce": self.open_orders_nonce,
            "series_expiry": self.series_expiry,
            "series_expiry_padding": self.series_expiry_padding,
            "product_ledgers": list(map(lambda item: item.to_json(), self.product_ledgers)),
            "product_ledgers_padding": list(map(lambda item: item.to_json(), self.product_ledgers_padding)),
            "perp_product_ledger": self.perp_product_ledger.to_json(),
            "rebalance_amount": self.rebalance_amount,
            "asset": self.asset.to_json(),
            "account_type": self.account_type.to_json(),
            "last_funding_delta": self.last_funding_delta.to_json(),
            "delegated_pubkey": str(self.delegated_pubkey),
            "rebate_rebalance_amount": self.rebate_rebalance_amount,
            "padding": self.padding,
        }

    @classmethod
    def from_json(cls, obj: MarginAccountJSON) -> "MarginAccount":
        return cls(
            authority=Pubkey.from_string(obj["authority"]),
            nonce=obj["nonce"],
            balance=obj["balance"],
            force_cancel_flag=obj["force_cancel_flag"],
            open_orders_nonce=obj["open_orders_nonce"],
            series_expiry=obj["series_expiry"],
            series_expiry_padding=obj["series_expiry_padding"],
            product_ledgers=list(
                map(
                    lambda item: types.product_ledger.ProductLedger.from_json(item),
                    obj["product_ledgers"],
                )
            ),
            product_ledgers_padding=list(
                map(
                    lambda item: types.product_ledger.ProductLedger.from_json(item),
                    obj["product_ledgers_padding"],
                )
            ),
            perp_product_ledger=types.product_ledger.ProductLedger.from_json(obj["perp_product_ledger"]),
            rebalance_amount=obj["rebalance_amount"],
            asset=types.asset.from_json(obj["asset"]),
            account_type=types.margin_account_type.from_json(obj["account_type"]),
            last_funding_delta=types.anchor_decimal.AnchorDecimal.from_json(obj["last_funding_delta"]),
            delegated_pubkey=Pubkey.from_string(obj["delegated_pubkey"]),
            rebate_rebalance_amount=obj["rebate_rebalance_amount"],
            padding=obj["padding"],
        )



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from anchorpy.borsh_extension import EnumForCodegen


class UninitializedJSON(typing.TypedDict):
    kind: typing.Literal["Uninitialized"]


class LessThanOrEqualJSON(typing.TypedDict):
    kind: typing.Literal["LessThanOrEqual"]


class GreaterThanOrEqualJSON(typing.TypedDict):
    kind: typing.Literal["GreaterThanOrEqual"]


@dataclass
class Uninitialized:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "Uninitialized"

    @classmethod
    def to_json(cls) -> UninitializedJSON:
        return UninitializedJSON(
            kind="Uninitialized",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Uninitialized": {},
        }


@dataclass
class LessThanOrEqual:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "LessThanOrEqual"

    @classmethod
    def to_json(cls) -> LessThanOrEqualJSON:
        return LessThanOrEqualJSON(
            kind="LessThanOrEqual",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "LessThanOrEqual": {},
        }


@dataclass
class GreaterThanOrEqual:
    discriminator: typing.ClassVar = 2
    kind: typing.ClassVar = "GreaterThanOrEqual"

    @classmethod
    def to_json(cls) -> GreaterThanOrEqualJSON:
        return GreaterThanOrEqualJSON(
            kind="GreaterThanOrEqual",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "GreaterThanOrEqual": {},
        }


TriggerDirectionKind = typing.Union[Uninitialized, LessThanOrEqual, GreaterThanOrEqual]
TriggerDirectionJSON = typing.Union[UninitializedJSON, LessThanOrEqualJSON, GreaterThanOrEqualJSON]


def from_decoded(obj: dict) -> TriggerDirectionKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "Uninitialized" in obj:
        return Uninitialized()
    if "LessThanOrEqual" in obj:
        return LessThanOrEqual()
    if "GreaterThanOrEqual" in obj:
        return GreaterThanOrEqual()
    raise ValueError("Invalid enum object")


def from_json(obj: TriggerDirectionJSON) -> TriggerDirectionKind:
    if obj["kind"] == "Uninitialized":
        return Uninitialized()
    if obj["kind"] == "LessThanOrEqual":
        return LessThanOrEqual()
    if obj["kind"] == "GreaterThanOrEqual":
        return GreaterThanOrEqual()
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen(
    "Uninitialized" / borsh.CStruct(),
    "LessThanOrEqual" / borsh.CStruct(),
    "GreaterThanOrEqual" / borsh.CStruct(),
)



from __future__ import annotations

import typing

import borsh_construct as borsh
from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey
from spl.token.constants import TOKEN_PROGRAM_ID

from .. import types
from ..program_id import PROGRAM_ID


class TreasuryMovementArgs(typing.TypedDict):
    treasury_movement_type: types.treasury_movement_type.TreasuryMovementTypeKind
    amount: int


layout = borsh.CStruct("treasury_movement_type" / types.treasury_movement_type.layout, "amount" / borsh.U64)


class TreasuryMovementAccounts(typing.TypedDict):
    state: Pubkey
    insurance_vault: Pubkey
    treasury_wallet: Pubkey
    referrals_rewards_wallet: Pubkey
    admin: Pubkey


def treasury_movement(
    args: TreasuryMovementArgs,
    accounts: TreasuryMovementAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["state"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["insurance_vault"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["treasury_wallet"], is_signer=False, is_writable=True),
        AccountMeta(
            pubkey=accounts["referrals_rewards_wallet"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(pubkey=TOKEN_PROGRAM_ID, is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["admin"], is_signer=True, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b'\x01"\xf2i\xd7\xd3\x9d\x12'
    encoded_args = layout.build(
        {
            "treasury_movement_type": args["treasury_movement_type"].to_encodable(),
            "amount": args["amount"],
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations

import typing

import borsh_construct as borsh
from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey
from solders.system_program import ID as SYS_PROGRAM_ID

from ..program_id import PROGRAM_ID


class InitializeInsuranceDepositAccountArgs(typing.TypedDict):
    nonce: int


layout = borsh.CStruct("nonce" / borsh.U8)


class InitializeInsuranceDepositAccountAccounts(typing.TypedDict):
    insurance_deposit_account: Pubkey
    authority: Pubkey
    payer: Pubkey
    whitelist_insurance_account: Pubkey


def initialize_insurance_deposit_account(
    args: InitializeInsuranceDepositAccountArgs,
    accounts: InitializeInsuranceDepositAccountAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(
            pubkey=accounts["insurance_deposit_account"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(pubkey=accounts["authority"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["payer"], is_signer=True, is_writable=True),
        AccountMeta(pubkey=SYS_PROGRAM_ID, is_signer=False, is_writable=False),
        AccountMeta(
            pubkey=accounts["whitelist_insurance_account"],
            is_signer=False,
            is_writable=False,
        ),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"U\xa3ry\x8b\xa7)%"
    encoded_args = layout.build(
        {
            "nonce": args["nonce"],
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations

import typing

import borsh_construct as borsh
from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey

from ..program_id import PROGRAM_ID


class PruneExpiredTifOrdersV2Args(typing.TypedDict):
    limit: int


layout = borsh.CStruct("limit" / borsh.U16)


class PruneExpiredTifOrdersV2Accounts(typing.TypedDict):
    dex_program: Pubkey
    state: Pubkey
    serum_authority: Pubkey
    market: Pubkey
    bids: Pubkey
    asks: Pubkey
    event_queue: Pubkey


def prune_expired_tif_orders_v2(
    args: PruneExpiredTifOrdersV2Args,
    accounts: PruneExpiredTifOrdersV2Accounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["dex_program"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["state"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["serum_authority"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["market"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["bids"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["asks"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["event_queue"], is_signer=False, is_writable=True),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xcc%|9\x9e\xd3\xb2\xd1"
    encoded_args = layout.build(
        {
            "limit": args["limit"],
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from anchorpy.borsh_extension import EnumForCodegen


class MarginAccountJSON(typing.TypedDict):
    kind: typing.Literal["MarginAccount"]


class CrossMarginAccountJSON(typing.TypedDict):
    kind: typing.Literal["CrossMarginAccount"]


@dataclass
class MarginAccount:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "MarginAccount"

    @classmethod
    def to_json(cls) -> MarginAccountJSON:
        return MarginAccountJSON(
            kind="MarginAccount",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "MarginAccount": {},
        }


@dataclass
class CrossMarginAccount:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "CrossMarginAccount"

    @classmethod
    def to_json(cls) -> CrossMarginAccountJSON:
        return CrossMarginAccountJSON(
            kind="CrossMarginAccount",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "CrossMarginAccount": {},
        }


TraitTypeKind = typing.Union[MarginAccount, CrossMarginAccount]
TraitTypeJSON = typing.Union[MarginAccountJSON, CrossMarginAccountJSON]


def from_decoded(obj: dict) -> TraitTypeKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "MarginAccount" in obj:
        return MarginAccount()
    if "CrossMarginAccount" in obj:
        return CrossMarginAccount()
    raise ValueError("Invalid enum object")


def from_json(obj: TraitTypeJSON) -> TraitTypeKind:
    if obj["kind"] == "MarginAccount":
        return MarginAccount()
    if obj["kind"] == "CrossMarginAccount":
        return CrossMarginAccount()
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen("MarginAccount" / borsh.CStruct(), "CrossMarginAccount" / borsh.CStruct())



from dataclasses import dataclass, field
from enum import Enum, IntEnum
from typing import Optional

from solders.hash import Hash
from solders.pubkey import Pubkey

from zetamarkets_py.zeta_client.types import (
    asset,
    order_type,
    self_trade_behavior_zeta,
    side,
)


class Asset(Enum):
    """Enum class for different types of assets."""

    SOL = "SOL"
    BTC = "BTC"
    ETH = "ETH"
    APT = "APT"
    ARB = "ARB"
    BNB = "BNB"
    PYTH = "PYTH"
    TIA = "TIA"
    JTO = "JTO"
    ONEMBONK = "ONEMBONK"
    SEI = "SEI"
    JUP = "JUP"
    DYM = "DYM"
    STRK = "STRK"
    WIF = "WIF"
    RNDR = "RNDR"
    TNSR = "TNSR"
    POPCAT = "POPCAT"
    EIGEN = "EIGEN"
    DBR = "DBR"
    GOAT = "GOAT"

    def to_index(self):
        """Converts the asset to its corresponding index."""
        members = list(self.__class__.__members__.values())
        return members.index(self)

    @classmethod
    def from_index(cls, index: int):
        """Returns the asset corresponding to the given index."""
        members = list(Asset)
        return members[index]

    def to_program_type(self):
        """Converts the asset to its corresponding program type."""
        return asset.from_decoded({self.name: self.to_index()})

    @staticmethod
    def all():
        """Returns a list of all assets."""
        return list(Asset)

    def __str__(self) -> str:
        """Returns the name of the asset."""
        return self.name


class Network(Enum):
    """Enum class for different types of networks."""

    DEVNET = "devnet"
    TESTNET = "testnet"
    MAINNET = "mainnet_beta"

    def __str__(self) -> str:
        """Returns the name of the network."""
        return self.name


class OrderType(IntEnum):
    """Enum class for different types of orders."""

    Limit = 0
    PostOnly = 1
    FillOrKill = 2
    ImmediateOrCancel = 3
    PostOnlySlide = 4

    def to_program_type(self):
        """Converts the order type to its corresponding program type."""
        return order_type.from_decoded({self.name: self.value})

    @classmethod
    def from_index(cls, index: int):
        """Returns the order type corresponding to the given index."""
        members = list(OrderType)
        return members[index]

    def __str__(self) -> str:
        """Returns the name of the order type."""
        return self.name


class SelfTradeBehaviorZeta(IntEnum):
    """Enum class for different types of orders."""

    CancelProvide = 0
    AbortTransaction = 1

    def to_program_type(self):
        """Converts the order type to its corresponding program type."""
        return self_trade_behavior_zeta.from_decoded({self.name: self.value})

    @classmethod
    def from_index(cls, index: int):
        """Returns the order type corresponding to the given index."""
        members = list(SelfTradeBehaviorZeta)
        return members[index]

    def __str__(self) -> str:
        """Returns the name of the order type."""
        return self.name


class Side(Enum):
    """Enum class for different sides of the orderbook to trade."""

    Uninitialized = 0
    Bid = 1
    Ask = 2

    def to_program_type(self) -> side.SideKind:
        """Converts the side to its corresponding program type."""
        return side.from_decoded({self.name: self.value})

    @classmethod
    def from_index(cls, index: int):
        """Returns the side corresponding to the given index."""
        members = list(Side)
        return members[index]

    def __str__(self) -> str:
        """Returns the name of the side."""
        return self.name


class SelfTradeBehavior(Enum):
    """Enum class for different types of self trade behaviors."""

    DecrementTake = 0
    CancelProvide = 1
    AbortTransaction = 2

    def __str__(self) -> str:
        """Returns the name of the self trade behavior."""
        return self.name


class OrderCompleteType(Enum):
    """Enum class for different types of order completion statuses."""

    Cancel = 0
    Fill = 1
    Booted = 2

    @classmethod
    def from_index(cls, index: int):
        """Returns the order completion status corresponding to the given index."""
        members = list(OrderCompleteType)
        return members[index]

    def __str__(self) -> str:
        """Returns the name of the order completion status."""
        return self.name


@dataclass
class TIFOptions:
    """Data class for Time in Force options."""

    expiry_offset: Optional[int] = None
    expiry_ts: Optional[int] = None


@dataclass
class OrderOptions:
    """Data class for order options."""

    expiry_ts: Optional[int] = None
    client_order_id: Optional[int] = None
    blockhash: Optional[Hash] = None
    order_type: OrderType = OrderType.Limit
    tag: str = "SDK"
    self_trade_behavior: Optional[SelfTradeBehaviorZeta] = None


@dataclass
class OrderInfo:
    """Data class for order information."""

    price: float
    size: float


@dataclass
class Order:
    """Data class for order details."""

    order_id: int
    client_id: int
    open_order_address: Pubkey
    open_order_slot: int
    fee_tier: int
    info: OrderInfo
    side: Side
    tif_offset: int


@dataclass
class FilledOrder:
    """Data class for filled order details."""

    order_id: int
    side: Side
    price: float
    size: float
    fee_cost: int


@dataclass
class OrderArgs:
    """Data class for order arguments."""

    price: float
    size: float
    side: Side
    order_opts: OrderOptions = field(default_factory=OrderOptions)


@dataclass
class MultiOrderArgs:
    """Data class for multi order arguments."""

    price: float
    size: float
    client_order_id: Optional[int] = None
    expiry_ts: Optional[int] = None


class AnchorDecimal:
    flags: int
    hi: int
    lo: int
    mid: int


class Decimal:
    def __init__(self, flags: int, hi: int, lo: int, mid: int):
        self._flags = flags
        self._hi = hi
        self._lo = lo
        self._mid = mid
        self.SCALE_MASK = 0x00FF0000
        self.SCALE_SHIFT = 16
        self.SIGN_MASK = 0x80000000

    @classmethod
    def from_anchor_decimal(cls, decimal: AnchorDecimal):
        return cls(decimal.flags, decimal.hi, decimal.lo, decimal.mid)

    def scale(self) -> int:
        return (self._flags & self.SCALE_MASK) >> self.SCALE_SHIFT

    def is_sign_negative(self) -> bool:
        return (self._flags & self.SIGN_MASK) != 0

    def is_sign_positive(self) -> bool:
        return (self._flags & self.SIGN_MASK) == 0

    def to_float(self):
        scale = self.scale()
        if scale == 0:
            raise ValueError("Scale 0 is not handled.")

        bytes_ = bytes(
            [
                (self._hi >> 24) & 0xFF,
                (self._hi >> 16) & 0xFF,
                (self._hi >> 8) & 0xFF,
                self._hi & 0xFF,
                (self._mid >> 24) & 0xFF,
                (self._mid >> 16) & 0xFF,
                (self._mid >> 8) & 0xFF,
                self._mid & 0xFF,
                (self._lo >> 24) & 0xFF,
                (self._lo >> 16) & 0xFF,
                (self._lo >> 8) & 0xFF,
                self._lo & 0xFF,
            ]
        )

        return (-1 * int.from_bytes(bytes_, "big") if self.is_sign_negative() else int.from_bytes(bytes_, "big")) / (
            10**scale
        )

    def is_unset(self) -> bool:
        return self._hi == 0 and self._mid == 0 and self._lo == 0 and self._flags == 0



import typing
from dataclasses import dataclass

import borsh_construct as borsh
from anchorpy.borsh_extension import BorshPubkey
from anchorpy.coder.accounts import ACCOUNT_DISCRIMINATOR_SIZE
from anchorpy.error import AccountInvalidDiscriminator
from anchorpy.utils.rpc import get_multiple_accounts
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
from solders.pubkey import Pubkey

from ..program_id import PROGRAM_ID


class ReferrerAccountJSON(typing.TypedDict):
    nonce: int
    has_alias: bool
    referrer: str
    pending_rewards: int
    claimed_rewards: int


@dataclass
class ReferrerAccount:
    discriminator: typing.ClassVar = b"0\x13\xa06L\xdcF\t"
    layout: typing.ClassVar = borsh.CStruct(
        "nonce" / borsh.U8,
        "has_alias" / borsh.Bool,
        "referrer" / BorshPubkey,
        "pending_rewards" / borsh.U64,
        "claimed_rewards" / borsh.U64,
    )
    nonce: int
    has_alias: bool
    referrer: Pubkey
    pending_rewards: int
    claimed_rewards: int

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        address: Pubkey,
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.Optional["ReferrerAccount"]:
        resp = await conn.get_account_info(address, commitment=commitment)
        info = resp.value
        if info is None:
            return None
        if info.owner != program_id:
            raise ValueError("Account does not belong to this program")
        bytes_data = info.data
        return cls.decode(bytes_data)

    @classmethod
    async def fetch_multiple(
        cls,
        conn: AsyncClient,
        addresses: list[Pubkey],
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.List[typing.Optional["ReferrerAccount"]]:
        infos = await get_multiple_accounts(conn, addresses, commitment=commitment)
        res: typing.List[typing.Optional["ReferrerAccount"]] = []
        for info in infos:
            if info is None:
                res.append(None)
                continue
            if info.account.owner != program_id:
                raise ValueError("Account does not belong to this program")
            res.append(cls.decode(info.account.data))
        return res

    @classmethod
    def decode(cls, data: bytes) -> "ReferrerAccount":
        if data[:ACCOUNT_DISCRIMINATOR_SIZE] != cls.discriminator:
            raise AccountInvalidDiscriminator("The discriminator for this account is invalid")
        dec = ReferrerAccount.layout.parse(data[ACCOUNT_DISCRIMINATOR_SIZE:])
        return cls(
            nonce=dec.nonce,
            has_alias=dec.has_alias,
            referrer=dec.referrer,
            pending_rewards=dec.pending_rewards,
            claimed_rewards=dec.claimed_rewards,
        )

    def to_json(self) -> ReferrerAccountJSON:
        return {
            "nonce": self.nonce,
            "has_alias": self.has_alias,
            "referrer": str(self.referrer),
            "pending_rewards": self.pending_rewards,
            "claimed_rewards": self.claimed_rewards,
        }

    @classmethod
    def from_json(cls, obj: ReferrerAccountJSON) -> "ReferrerAccount":
        return cls(
            nonce=obj["nonce"],
            has_alias=obj["has_alias"],
            referrer=Pubkey.from_string(obj["referrer"]),
            pending_rewards=obj["pending_rewards"],
            claimed_rewards=obj["claimed_rewards"],
        )



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from construct import Container


class InitializeStateArgsJSON(typing.TypedDict):
    state_nonce: int
    serum_nonce: int
    mint_auth_nonce: int
    strike_initialization_threshold_seconds: int
    pricing_frequency_seconds: int
    liquidator_liquidation_percentage: int
    insurance_vault_liquidation_percentage: int
    native_deposit_limit: int
    expiration_threshold_seconds: int
    position_movement_fee_bps: int
    margin_concession_percentage: int
    max_perp_delta_age_seconds: int
    native_withdraw_limit: int
    withdraw_limit_epoch_seconds: int
    native_open_interest_limit: int


@dataclass
class InitializeStateArgs:
    layout: typing.ClassVar = borsh.CStruct(
        "state_nonce" / borsh.U8,
        "serum_nonce" / borsh.U8,
        "mint_auth_nonce" / borsh.U8,
        "strike_initialization_threshold_seconds" / borsh.U32,
        "pricing_frequency_seconds" / borsh.U32,
        "liquidator_liquidation_percentage" / borsh.U32,
        "insurance_vault_liquidation_percentage" / borsh.U32,
        "native_deposit_limit" / borsh.U64,
        "expiration_threshold_seconds" / borsh.U32,
        "position_movement_fee_bps" / borsh.U8,
        "margin_concession_percentage" / borsh.U8,
        "max_perp_delta_age_seconds" / borsh.U16,
        "native_withdraw_limit" / borsh.U64,
        "withdraw_limit_epoch_seconds" / borsh.U32,
        "native_open_interest_limit" / borsh.U64,
    )
    state_nonce: int
    serum_nonce: int
    mint_auth_nonce: int
    strike_initialization_threshold_seconds: int
    pricing_frequency_seconds: int
    liquidator_liquidation_percentage: int
    insurance_vault_liquidation_percentage: int
    native_deposit_limit: int
    expiration_threshold_seconds: int
    position_movement_fee_bps: int
    margin_concession_percentage: int
    max_perp_delta_age_seconds: int
    native_withdraw_limit: int
    withdraw_limit_epoch_seconds: int
    native_open_interest_limit: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "InitializeStateArgs":
        return cls(
            state_nonce=obj.state_nonce,
            serum_nonce=obj.serum_nonce,
            mint_auth_nonce=obj.mint_auth_nonce,
            strike_initialization_threshold_seconds=obj.strike_initialization_threshold_seconds,
            pricing_frequency_seconds=obj.pricing_frequency_seconds,
            liquidator_liquidation_percentage=obj.liquidator_liquidation_percentage,
            insurance_vault_liquidation_percentage=obj.insurance_vault_liquidation_percentage,
            native_deposit_limit=obj.native_deposit_limit,
            expiration_threshold_seconds=obj.expiration_threshold_seconds,
            position_movement_fee_bps=obj.position_movement_fee_bps,
            margin_concession_percentage=obj.margin_concession_percentage,
            max_perp_delta_age_seconds=obj.max_perp_delta_age_seconds,
            native_withdraw_limit=obj.native_withdraw_limit,
            withdraw_limit_epoch_seconds=obj.withdraw_limit_epoch_seconds,
            native_open_interest_limit=obj.native_open_interest_limit,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "state_nonce": self.state_nonce,
            "serum_nonce": self.serum_nonce,
            "mint_auth_nonce": self.mint_auth_nonce,
            "strike_initialization_threshold_seconds": self.strike_initialization_threshold_seconds,
            "pricing_frequency_seconds": self.pricing_frequency_seconds,
            "liquidator_liquidation_percentage": self.liquidator_liquidation_percentage,
            "insurance_vault_liquidation_percentage": self.insurance_vault_liquidation_percentage,
            "native_deposit_limit": self.native_deposit_limit,
            "expiration_threshold_seconds": self.expiration_threshold_seconds,
            "position_movement_fee_bps": self.position_movement_fee_bps,
            "margin_concession_percentage": self.margin_concession_percentage,
            "max_perp_delta_age_seconds": self.max_perp_delta_age_seconds,
            "native_withdraw_limit": self.native_withdraw_limit,
            "withdraw_limit_epoch_seconds": self.withdraw_limit_epoch_seconds,
            "native_open_interest_limit": self.native_open_interest_limit,
        }

    def to_json(self) -> InitializeStateArgsJSON:
        return {
            "state_nonce": self.state_nonce,
            "serum_nonce": self.serum_nonce,
            "mint_auth_nonce": self.mint_auth_nonce,
            "strike_initialization_threshold_seconds": self.strike_initialization_threshold_seconds,
            "pricing_frequency_seconds": self.pricing_frequency_seconds,
            "liquidator_liquidation_percentage": self.liquidator_liquidation_percentage,
            "insurance_vault_liquidation_percentage": self.insurance_vault_liquidation_percentage,
            "native_deposit_limit": self.native_deposit_limit,
            "expiration_threshold_seconds": self.expiration_threshold_seconds,
            "position_movement_fee_bps": self.position_movement_fee_bps,
            "margin_concession_percentage": self.margin_concession_percentage,
            "max_perp_delta_age_seconds": self.max_perp_delta_age_seconds,
            "native_withdraw_limit": self.native_withdraw_limit,
            "withdraw_limit_epoch_seconds": self.withdraw_limit_epoch_seconds,
            "native_open_interest_limit": self.native_open_interest_limit,
        }

    @classmethod
    def from_json(cls, obj: InitializeStateArgsJSON) -> "InitializeStateArgs":
        return cls(
            state_nonce=obj["state_nonce"],
            serum_nonce=obj["serum_nonce"],
            mint_auth_nonce=obj["mint_auth_nonce"],
            strike_initialization_threshold_seconds=obj["strike_initialization_threshold_seconds"],
            pricing_frequency_seconds=obj["pricing_frequency_seconds"],
            liquidator_liquidation_percentage=obj["liquidator_liquidation_percentage"],
            insurance_vault_liquidation_percentage=obj["insurance_vault_liquidation_percentage"],
            native_deposit_limit=obj["native_deposit_limit"],
            expiration_threshold_seconds=obj["expiration_threshold_seconds"],
            position_movement_fee_bps=obj["position_movement_fee_bps"],
            margin_concession_percentage=obj["margin_concession_percentage"],
            max_perp_delta_age_seconds=obj["max_perp_delta_age_seconds"],
            native_withdraw_limit=obj["native_withdraw_limit"],
            withdraw_limit_epoch_seconds=obj["withdraw_limit_epoch_seconds"],
            native_open_interest_limit=obj["native_open_interest_limit"],
        )



from __future__ import annotations

import typing
from dataclasses import dataclass

import borsh_construct as borsh
from anchorpy.borsh_extension import EnumForCodegen


class UndefinedJSON(typing.TypedDict):
    kind: typing.Literal["Undefined"]


class ToTreasuryFromInsuranceJSON(typing.TypedDict):
    kind: typing.Literal["ToTreasuryFromInsurance"]


class ToInsuranceFromTreasuryJSON(typing.TypedDict):
    kind: typing.Literal["ToInsuranceFromTreasury"]


class ToTreasuryFromReferralsRewardsJSON(typing.TypedDict):
    kind: typing.Literal["ToTreasuryFromReferralsRewards"]


class ToReferralsRewardsFromTreasuryJSON(typing.TypedDict):
    kind: typing.Literal["ToReferralsRewardsFromTreasury"]


@dataclass
class Undefined:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "Undefined"

    @classmethod
    def to_json(cls) -> UndefinedJSON:
        return UndefinedJSON(
            kind="Undefined",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Undefined": {},
        }


@dataclass
class ToTreasuryFromInsurance:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "ToTreasuryFromInsurance"

    @classmethod
    def to_json(cls) -> ToTreasuryFromInsuranceJSON:
        return ToTreasuryFromInsuranceJSON(
            kind="ToTreasuryFromInsurance",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "ToTreasuryFromInsurance": {},
        }


@dataclass
class ToInsuranceFromTreasury:
    discriminator: typing.ClassVar = 2
    kind: typing.ClassVar = "ToInsuranceFromTreasury"

    @classmethod
    def to_json(cls) -> ToInsuranceFromTreasuryJSON:
        return ToInsuranceFromTreasuryJSON(
            kind="ToInsuranceFromTreasury",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "ToInsuranceFromTreasury": {},
        }


@dataclass
class ToTreasuryFromReferralsRewards:
    discriminator: typing.ClassVar = 3
    kind: typing.ClassVar = "ToTreasuryFromReferralsRewards"

    @classmethod
    def to_json(cls) -> ToTreasuryFromReferralsRewardsJSON:
        return ToTreasuryFromReferralsRewardsJSON(
            kind="ToTreasuryFromReferralsRewards",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "ToTreasuryFromReferralsRewards": {},
        }


@dataclass
class ToReferralsRewardsFromTreasury:
    discriminator: typing.ClassVar = 4
    kind: typing.ClassVar = "ToReferralsRewardsFromTreasury"

    @classmethod
    def to_json(cls) -> ToReferralsRewardsFromTreasuryJSON:
        return ToReferralsRewardsFromTreasuryJSON(
            kind="ToReferralsRewardsFromTreasury",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "ToReferralsRewardsFromTreasury": {},
        }


TreasuryMovementTypeKind = typing.Union[
    Undefined,
    ToTreasuryFromInsurance,
    ToInsuranceFromTreasury,
    ToTreasuryFromReferralsRewards,
    ToReferralsRewardsFromTreasury,
]
TreasuryMovementTypeJSON = typing.Union[
    UndefinedJSON,
    ToTreasuryFromInsuranceJSON,
    ToInsuranceFromTreasuryJSON,
    ToTreasuryFromReferralsRewardsJSON,
    ToReferralsRewardsFromTreasuryJSON,
]


def from_decoded(obj: dict) -> TreasuryMovementTypeKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "Undefined" in obj:
        return Undefined()
    if "ToTreasuryFromInsurance" in obj:
        return ToTreasuryFromInsurance()
    if "ToInsuranceFromTreasury" in obj:
        return ToInsuranceFromTreasury()
    if "ToTreasuryFromReferralsRewards" in obj:
        return ToTreasuryFromReferralsRewards()
    if "ToReferralsRewardsFromTreasury" in obj:
        return ToReferralsRewardsFromTreasury()
    raise ValueError("Invalid enum object")


def from_json(obj: TreasuryMovementTypeJSON) -> TreasuryMovementTypeKind:
    if obj["kind"] == "Undefined":
        return Undefined()
    if obj["kind"] == "ToTreasuryFromInsurance":
        return ToTreasuryFromInsurance()
    if obj["kind"] == "ToInsuranceFromTreasury":
        return ToInsuranceFromTreasury()
    if obj["kind"] == "ToTreasuryFromReferralsRewards":
        return ToTreasuryFromReferralsRewards()
    if obj["kind"] == "ToReferralsRewardsFromTreasury":
        return ToReferralsRewardsFromTreasury()
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen(
    "Undefined" / borsh.CStruct(),
    "ToTreasuryFromInsurance" / borsh.CStruct(),
    "ToInsuranceFromTreasury" / borsh.CStruct(),
    "ToTreasuryFromReferralsRewards" / borsh.CStruct(),
    "ToReferralsRewardsFromTreasury" / borsh.CStruct(),
)



from __future__ import annotations

import typing

import borsh_construct as borsh
from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey

from ..program_id import PROGRAM_ID


class CancelTriggerOrderV2Args(typing.TypedDict):
    trigger_order_bit: int


layout = borsh.CStruct("trigger_order_bit" / borsh.U8)


class CancelTriggerOrderV2Accounts(typing.TypedDict):
    authority: Pubkey
    trigger_order: Pubkey
    margin_account: Pubkey


def cancel_trigger_order_v2(
    args: CancelTriggerOrderV2Args,
    accounts: CancelTriggerOrderV2Accounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["authority"], is_signer=True, is_writable=True),
        AccountMeta(pubkey=accounts["trigger_order"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["margin_account"], is_signer=False, is_writable=True),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xdfAC\x07\xbd\x03?\x8e"
    encoded_args = layout.build(
        {
            "trigger_order_bit": args["trigger_order_bit"],
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations

import typing

from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey

from ..program_id import PROGRAM_ID


class UpdateAdminAccounts(typing.TypedDict):
    state: Pubkey
    admin: Pubkey
    new_admin: Pubkey


def update_admin(
    accounts: UpdateAdminAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["state"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["admin"], is_signer=True, is_writable=True),
        AccountMeta(pubkey=accounts["new_admin"], is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xa1\xb0(\xd5<\xb8\xb3\xe4"
    encoded_args = b""
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)
